from .wing import Wing
import numpy as np


class Vertical(Wing):
    """
    Class containing all methods and parameters related to Vertical stabilizers.

    Subclass of wing, has all the same variables.
    """
    weight_averages = [.2, .4, .4] # [Raymer, Torenbeek, NASA] - weighted averages used for weight estimation
    def __init__(self, params):
        """
        Initialize vertical stabilizer with input parameters.

        params: <dict> list of parameters to edit
        """
        super().__init__(params)

        # Set sections for AVL and other class parameters
        self.avl_sections = [[self.xle, self.yle, self.zle, self.cr, 0],
                         [(self.xle + self.cr / 4) + self.span * np.tan(self.sweep) - self.ct / 4, self.yle, self.zle + self.span, self.ct, 0]]
        xle_tip = self.xle + self.span * np.tan(self.sweep) - .25 * self.ct
        self.sweep_le = np.arctan((xle_tip - self.xle) / self.span)
        self.Q = 1.1

    def raymer_weight(self, aircraft, wdg):
        """
        Set vertical stabilizer weight using Raymer method.

        :param object aircraft: aircraft object which this nacelle belongs to
        :param int wdg: gross takeoff weight

        :return: Raymer weight (lbs)
        :rtype: float
        """
        # Set method parameters
        ht = 0
        for comp in aircraft.aero_components.values():
            if comp.component_type == 'Horizontal':
                if comp.zle > (aircraft.aero_components['Fuselage'].height / 2):
                    ht = 1
        if aircraft.aircraft_type == 'transport':
            lt = (self.xle + .25 * self.cr) - \
                 (aircraft.aero_components['Main Wing'].xle + .25 * aircraft.aero_components['Main Wing'].cr)

            self.weight_raymer = .0026 * (
                        1 + ht) ** .225 * wdg ** .556 * aircraft.ultimate_load ** .536 * lt ** -.5 * \
                                 self.area ** .5 * lt ** .875 * np.cos(self.sweep) ** -1 * self.aspect_ratio ** .35 \
                                 * self.tc ** -.5
            return self.weight_raymer
        elif aircraft.aircraft_type == 'general_aviation':
            a = .073 * (1 + .2 * ht) * (aircraft.ultimate_load * wdg) ** .376
            b = aircraft.cruise_conditions.q ** .122 * self.area ** .873
            c = ((100 * self.tc) / np.cos(self.sweep)) ** -.49
            d = (self.aspect_ratio / (np.cos(self.sweep) ** 2)) ** .357
            e = self.taper ** .039

            self.weight_raymer = a * b * c * d * e
            return self.weight_raymer



    def torenbeek_weight(self, aircraft, wdg):
        """
        Set vertical stabilizer weight using Torenbeek Method.

        :param object aircraft: <object aircraft>  aircraft which the horizontal stabilizer belongs to
        :param int wdg: takeoff gross weight (lbs)

        :return: Torenbeek weight (lbs)
        :rtype: float
        """
        # check if horizontal is mounted on vertical
        horiz = False
        for comp in aircraft.aero_components.values():
            if comp.component_type == 'Horizontal':
                if comp.zle > (aircraft.aero_components['Fuselage'].height / 2):
                    horiz = True
                    horiz_comp = comp
                break

        if horiz:
            kv = 1 + .15 * ((horiz_comp.area * horiz_comp.zle) / (self.area * self.span))
        else:
            kv = 1

        dive_speed = aircraft.cruise_conditions.velocity * .592484
        a = kv * self.area * 3.81
        b = (self.area ** .2 * dive_speed) / (1000 * np.cos(self.sweep_mid) ** .5)
        self.weight_torenbeek = a * b ** -.281
        return self.weight_torenbeek

    def nasa_weight(self, aircraft, wdg):
        """
        Set vertical stablilizer weight using NASA FLOPS method.

        :param: object aircraft: <aircraft object> aircraft which the horizontal stabilizer belongs to
        :param: int wdg: takeoff gross weight (lbs)

        :return: NASA FLOPS weight (lbs)
        :rtype: float
        """
        ht = 0
        for comp in aircraft.aero_components.values():
            if comp.component_type == 'Horizontal':
                if comp.zle > (aircraft.aero_components['Fuselage'].height / 2):
                    ht = 1
        if aircraft.aircraft_type == 'transport':
            if self.yle == 0 and self.dihedral_deg == 90:
                n_vert = 1
            else:
                n_vert = 2
            self.weight_nasa = .32 * wdg ** .3 * (self.taper + .5) * n_vert ** .7 * self.area ** .85
            return self.weight_nasa
        elif aircraft.aircraft_type == 'general_aviation':
            Csweep = np.cos(self.sweep)
            a = .073 * (1 + .2 * ht) * (aircraft.ultimate_load * wdg) ** .376
            b = aircraft.cruise_conditions.q ** .122 * self.area ** .873
            c = (self.aspect_ratio / (Csweep ** 2)) ** .357
            d = ((100 * self.tc)/ Csweep) ** .49
            self.weight_nasa = a * b * (c / d)
            return self.weight_nasa

    def set_cg(self):
        """
        Sets center of gravity and moments of inertia for vertical stabilizer
        """
        a = self.ct
        b = self.cr
        semi_span = self.span * .5
        c = semi_span * np.tan(self.sweep_le)
        dihedral = self.dihedral

        self.cg = [self.xle + (2 * a * c + a ** 2 + c * b + a * b + b ** 2) / (3 * (a + b)),
                   self.yle,
                   self.zle + (semi_span * (2 * a + b) / (3 * (a + b)))]

        self.inertia = [x * self.weight for x in self.cg]

    def set_wave_drag(self, aircraft):
        """
        Set the wave drag of the vertical stabilizer to zero.

        :param object aircraft: <aircraft object> aircraft which the horizontal stabilizer belongs to
        :return: 0
        :rytpe: int
        """
        self.cdw = 0
        return 0


import logging
import pandas as pd
from openpyxl import load_workbook
from openpyxl.styles import Font, Alignment
from .flight_conditions import FlightConditions
import numpy as np
logger = logging.getLogger(__name__)

class engine:
    """
    Contains all analyses and variables required to analyze propulsion performance
    """
    engine_type = ''
    altitude_ref = []  # Altitudes used for the sfc and thrust tables
    mach_ref = []  # Mach number used for the sfc and thrust tables
    thrust_input = []  # Input values for thrust
    sfc_input = []  # Input values for sfc
    components = {}

    def __init__(self, n_engines=2, max_thrust=0, sfc=0):
        """
        Initializes the number of engines with the specified number of engines.

        :param int n_engines: Number of engines
        """
        self.n_engines = n_engines  # Number of engines
        self.max_thrust = max_thrust
        self.sfc = sfc

    def analyze_performance(self, height, mach, thrust_required=None):
        return


class turbofan(engine):
    """
    Represents a generic, commercial aviation turbofan engine

    Scales results based off the values for the cfm56_7b24 (737-800's engine)
    """

    # Values used to scale the standard performance data
    thrust_sea_level = None
    thrust_cruise = None
    sfc_sea_level = None
    sfc_cruise = None
    engine_type = None
    engine_data_file = None

    engine_type = 'turbofan'
    # Basic cfm56-7b values
    thrust_input = [[24200, 21016, 19105, 17004, 15284, 13819, 12418, 10508, 8152, 5413],
                    [16004, 14597, 13954, 12293, 11335, 10505, 9932, 9165, 8271, 7185],
                    [9788, 9074, 8425, 7842, 7326, 6909, 6685, 6426, 6153, 5585],
                    [8155, 7522, 6915, 6459, 6165, 6002, 5937, 5741, 5480, 5121],
                    [6214, 5750, 5320, 4988, 4757, 4625, 4515, 4363, 4178, 3932],
                    [4054, 3516, 3113, 2819, 2605, 2443, 2305, 2162, 1987, 1752]]

    sfc_input = [[0.370, 0.416, 0.474, 0.503, 0.624, 0.694, 0.786, 0.948, 1.214, 1.769],
                 [0.362, 0.382, 0.413, 0.467, 0.527, 0.592, 0.674, 0.760, 0.869, 1.046],
                 [0.294, 0.333, 0.367, 0.408, 0.449, 0.505, 0.541, 0.615, 0.670, 0.769],
                 [0.291, 0.324, 0.353, 0.390, 0.425, 0.462, 0.509, 0.564, 0.627, 0.703],
                 [0.273, 0.333, 0.359, 0.383, 0.419, 0.466, 0.498, 0.563, 0.614, 0.661],
                 [0.224, 0.260, 0.296, 0.331, 0.367, 0.406, 0.448, 0.495, 0.548, 0.609]]

    mach_ref = [0, .1, .2, .3, .4, .5, .6, .7, .8, .9]
    altitude_ref = [0, 15000, 31000, 35000, 41000, 50000]

    def __init__(self, h_cruise, mach_cruise, n_engines=2, thrust_sea_level=None,
                 sfc_sea_level=None, thrust_cruise=None, sfc_cruise=None, engine_data_file=None):
        """
        Scales the results based off the input conditions at both sea level and cruise

        inputs:
        :param float h_cruise: altitude at cruise (ft)
        :param float mach_cruise: mach number at cruise
        :param int n_engines: Number of engines
        :param float thrust_sea_level: maximum thrust at sea level, lbs
        :param float sfc_sea_level: specific fuel consumption at sea level
        :param float thrust_cruise: maximum thrust at cruise conditions, lbs
        :param float sfc_cruise: specific fuel consuption at cruise, 1/hrs
        """
        super().__init__(n_engines)
        if engine_data_file:
            self.load_data_file(engine_data_file)

        self.scale_performance(h_cruise, mach_cruise, n_engines=n_engines, thrust_sea_level=thrust_sea_level,
                               thrust_cruise=thrust_cruise,
                               sfc_cruise=sfc_cruise, sfc_sea_level=sfc_sea_level)

    def analyze_performance(self, height, mach, thrust_required=None):
        """
        Find the maximum available thrust and specific fuel consumption at an input altitude, mach number,
        and required thrust


        :param float height: Altitude (ft)
        :param float mach: Mach Number
        :param float thrust_required: Required thrust (lbs)


        :return float sfc: specific fuel consumption at input conditions (s^-1)
        :return float max_thrust: Maximum thrust at input conditions (lbs)

        """
        if mach > self.mach_ref[-1]:  # TODO edit interpolation to include mach and altitudes outside input list
            mach = self.mach_ref[-1] - .001

        # Find values to interpolate between
        for i in range(len(self.altitude_ref)):
            if height < self.altitude_ref[i]:
                if i > 0:
                    x_indx_low = i - 1
                    x_indx_high = i
                    x1 = self.altitude_ref[i - 1]
                    x2 = self.altitude_ref[i]
                else:
                    x_indx_low = 0
                    x_indx_high = 1
                    x1 = self.altitude_ref[0]
                    x2 = self.altitude_ref[1]
                break
        for i in range(len(self.mach_ref)):
            if mach < self.mach_ref[i]:
                y_indx_low = i - 1
                y_indx_high = i
                y1 = self.mach_ref[i - 1]
                y2 = self.mach_ref[i]
                break

        f11 = self.sfc_input[x_indx_low][y_indx_low]
        f12 = self.sfc_input[x_indx_low][y_indx_high]
        f21 = self.sfc_input[x_indx_high][y_indx_low]
        f22 = self.sfc_input[x_indx_high][y_indx_high]

        sfc = self._weighted_bilinear(x1, x2, y1, y2, f11, f12, f21, f22, height, mach)

        # Now find the max thrust
        f11 = self.thrust_input[x_indx_low][y_indx_low]
        f12 = self.thrust_input[x_indx_low][y_indx_high]
        f21 = self.thrust_input[x_indx_high][y_indx_low]
        f22 = self.thrust_input[x_indx_high][y_indx_high]

        max_thrust = self._weighted_bilinear(x1, x2, y1, y2, f11, f12, f21, f22, height, mach)
        max_thrust *= self.n_engines

        # Scale results for partial power (if required thrust is input)
        if thrust_required:
            thrust_percent = thrust_required / max_thrust
            # Raymer partial power correction
            if thrust_percent > .2:
                sfc *= .1 / thrust_percent + .24 / (thrust_percent) ** .8 + .66 * thrust_percent ** .8 + .1 * mach * (
                        1 / thrust_percent - thrust_percent)
            else:
                sfc *= 1.5

        return sfc, max_thrust

    def scale_performance(self, h_cruise, mach_cruise, n_engines=2, thrust_sea_level=None,
                          sfc_sea_level=None, thrust_cruise=None, sfc_cruise=None):

        if h_cruise > self.altitude_ref[-1]:
            # TODO make this work outside bounds
            logging.warning(
                f'Input cruise altitude is higher than the standard engine bounds ({self.altitude_ref[-1]} ft), could not scale engine data. Please select a different engine or lower the cruise altitude',
                UserWarning)
            return

        self.thrust_sea_level = thrust_sea_level
        self.thrust_cruise = thrust_cruise
        self.sfc_sea_level = sfc_sea_level
        self.sfc_cruise = sfc_cruise

        # Set scale values
        if thrust_sea_level:
            thrust_scale_sl = thrust_sea_level / self.thrust_input[0][0]
        else:
            thrust_scale_sl = 1

        if sfc_sea_level:
            sfc_scale_sl = sfc_sea_level / self.sfc_input[0][0]
        else:
            sfc_scale_sl = 1

        sfc_ref, thrust_cruise_ref = self.analyze_performance(h_cruise, mach_cruise)
        thrust_cruise_ref *= 1 / self.n_engines  # Single engine correction
        if thrust_cruise:
            thrust_scale_cruise = thrust_cruise / thrust_cruise_ref
        else:
            thrust_scale_cruise = 1

        if sfc_cruise:
            sfc_scale_cruise = sfc_cruise / sfc_ref
        else:
            sfc_scale_cruise = 1

        # Scale all values
        for i in range(len(self.thrust_input)):
            for j in range(len(self.thrust_input[i])):
                t_scaled_sea_level = self.thrust_input[i][j] * thrust_scale_sl
                t_scaled_cruise = self.thrust_input[i][j] * thrust_scale_cruise
                sfc_scaled_sl = self.sfc_input[i][j] * sfc_scale_sl
                sfc_scaled_cruise = self.sfc_input[i][j] * sfc_scale_cruise

                if thrust_sea_level and not thrust_cruise:
                    self.thrust_input[i][j] = t_scaled_sea_level
                elif thrust_cruise and not thrust_sea_level:
                    self.thrust_input[i][j] = t_scaled_cruise
                elif thrust_cruise and thrust_sea_level:
                    self.thrust_input[i][j] = (h_cruise - self.altitude_ref[i]) / h_cruise * t_scaled_sea_level + \
                                              self.altitude_ref[i] / h_cruise * t_scaled_cruise

                if sfc_sea_level and not sfc_cruise:
                    self.sfc_input[i][j] = sfc_scaled_sl
                elif sfc_cruise and not sfc_sea_level:
                    self.sfc_input[i][j] = sfc_scaled_cruise
                elif sfc_cruise and sfc_sea_level:
                    self.sfc_input[i][j] = (h_cruise - self.altitude_ref[i]) / h_cruise * sfc_scaled_sl + \
                                           self.altitude_ref[i] / h_cruise * sfc_scaled_cruise

        self.max_thrust = self.thrust_input[0][0]

    @staticmethod
    def _weighted_bilinear(x1, x2, y1, y2, f11, f12, f21, f22, x, y):
        """
        Performs weighted bilinear interpolation for a given set of points.

        :param float x1: Lower bound for the x-coordinate.
        :param float x2: Upper bound for the x-coordinate.
        :param float y1: Lower bound for the y-coordinate.
        :param float y2: Upper bound for the y-coordinate.
        :param float f11: Value at (x1, y1).
        :param float f12: Value at (x1, y2).
        :param float f21: Value at (x2, y1).
        :param float f22: Value at (x2, y2).
        :param float x: Target x-coordinate for interpolation.
        :param float y: Target y-coordinate for interpolation.

        :return: Interpolated value at (x, y).
        :rtype: float
        """
        # Weighted Bilinear interpolation
        w11 = ((x2 - x) * (y2 - y)) / ((x2 - x1) * (y2 - y1))
        w12 = ((x2 - x) * (y - y1)) / ((x2 - x1) * (y2 - y1))
        w21 = ((x - x1) * (y2 - y)) / ((x2 - x1) * (y2 - y1))
        w22 = ((x - x1) * (y - y1)) / ((x2 - x1) * (y2 - y1))

        f = w11 * f11 + w12 * f12 + w21 * f21 + w22 * f22
        return f

    def write_data_file(self, file_name):
        # Output data file to correct location
        thrust_df = pd.DataFrame(self.thrust_input, columns=self.mach_ref, index=self.altitude_ref)
        sfc_df = pd.DataFrame(self.sfc_input, columns=self.mach_ref, index=self.altitude_ref)
        with pd.ExcelWriter(file_name, engine="openpyxl") as writer:
            thrust_df.to_excel(writer, sheet_name="Thrust", startrow=2)
            sfc_df.to_excel(writer, sheet_name="sfc", startrow=2)

        # Fix formatting
        wb = load_workbook(file_name)

        col_headers = self.mach_ref
        row_headers = self.thrust_input
        for sheet_name in ["Thrust", "sfc"]:
            ws = wb[sheet_name]

            # Overall title (row 1, merged across all columns including row header column)
            max_col = len(col_headers) + 1  # +1 for row headers column
            ws.merge_cells(start_row=1, start_column=1, end_row=1, end_column=max_col)
            title_cell = ws.cell(row=1, column=1)
            if sheet_name == "Thrust":
                sheet_name += ' (lbs)'
            else:
                sheet_name += ' (lbm/hr)/lbs'
            title_cell.value = sheet_name
            title_cell.font = Font(size=16, bold=True)
            title_cell.alignment = Alignment(horizontal="center", vertical="center")

            # Column headers description (row 3, merged across data columns only, leaving row header column out)
            ws.merge_cells(start_row=2, start_column=2, end_row=2, end_column=max_col)
            col_desc_cell = ws.cell(row=2, column=2)
            col_desc_cell.value = 'Mach Number'
            col_desc_cell.alignment = Alignment(horizontal="center", vertical="center")
            col_desc_cell.font = Font(italic=True, size=11)

            # Row headers description (put near top-left, above row headers - cell A4)
            row_desc_cell = ws.cell(row=3, column=1)
            row_desc_cell.value = "Altitude (ft)"
            row_desc_cell.font = Font(italic=True, size=11)
            row_desc_cell.alignment = Alignment(horizontal="left", vertical="center")

            wb.save(file_name)

    def load_data_file(self, file_name):
        thrust_df = pd.read_excel(file_name, sheet_name="Thrust", skiprows=3, header=0, index_col=0)
        sfc_df = pd.read_excel(file_name, sheet_name="sfc", skiprows=3, header=0, index_col=0)

        self.altitude_ref = thrust_df.index.to_list()
        self.mach_ref = thrust_df.columns.tolist()
        self.thrust_input = thrust_df.values.tolist()
        self.sfc_input = sfc_df.values.tolist()


class propeller(engine):
    """
    Standard piston powered propeller engine
    """

    engine_type = 'propeller'

    def __init__(self, n_engines=1, horse_power=None, fuel_consumption_rate=None):
        super().__init__(n_engines)
        self.horse_power = horse_power
        self.fuel_consumption_rate = fuel_consumption_rate

    def analyze_performance(self, height, mach, thrust_required=None):
        nu_propeller = self.prop_efficiency(mach)

        # Flight conditions and velocity (ft/s)
        fc = FlightConditions(height, mach)
        V = mach * fc.a  # ft/s

        total_horse_power = self.n_engines * self.horse_power
        P_generated = total_horse_power * 550  # horsepower to ft-lb/s
        P_available = nu_propeller * P_generated

        fuel_density = 6.2  # lbs/gal (Avgas density adjusted slightly for temperature differences)
        total_fuel_consumption = self.n_engines * self.fuel_consumption_rate  # gal/hr
        fuel_mass_flow_rate_lb_per_hr = total_fuel_consumption * fuel_density  # lbs/hr

        # Handle very low or zero velocity case (static thrust approximation)
        if V <= 1e-3 or np.isnan(V):
            # Estimate static thrust as empirical factor * total horsepower
            static_thrust_factor = 2.5  # lbf per horsepower (typical GA propeller estimate)
            max_thrust = total_horse_power * static_thrust_factor

            sfc = fuel_mass_flow_rate_lb_per_hr / (max_thrust + 1e-6)  # Avoid div by zero
        else:
            max_thrust = P_available / V  # lbf
            # thrust calculated from power and velocity
            thrust = max_thrust
            sfc = fuel_mass_flow_rate_lb_per_hr / thrust  # lb/(lbf*hr)

        self.max_thrust = max_thrust
        return sfc, max_thrust

    # alernative method from section 3.1: https://www.fzt.haw-hamburg.de/pers/Scholz/transfer/Airport2030_TN_Propeller-Efficiency_13-08-12_SLZ.pdf
    @staticmethod
    def prop_efficiency(mach):
        nu_max = 0.85  # can adjust
        if mach == 0:
            nu_propeller = .8  # default? not sure what happens at 0 velocity
        elif 0 < mach <= 0.1:
            nu_propeller = 10 * mach * nu_max
        elif .1 < mach <= 0.7:
            nu_propeller = nu_max
        elif .7 < mach < 0.85:
            nu_propeller = 10 * mach * nu_max * (1 - (mach - 0.7) / 3)
        else:
            nu_propeller = max(0.0, 10 * mach * nu_max * (
                        1 - (mach - 0.7) / 3))  # prevent negative (just sets to 0 if negative)

        return nu_propeller



from PySide6.QtWidgets import QDialog, QFormLayout, QLineEdit, QLabel, QDialogButtonBox


class useful_load_edit(QDialog):
    """
    Handles useful load info
    """

    variables = {'pilots': 'Number',
                 'flight_attendants': 'Number',
                 'passengers': 'Number',
                 'fuel': 'Weight',
                 'cargo': 'Weight'}

    def __init__(self, parent, load_type):
        self.parent = parent
        super().__init__(parent)
        self.setWindowTitle('Useful Load Edit')

        form = QFormLayout()
        self.setLayout(form)

        type_edit = QLineEdit()
        type_edit.setText(load_type)
        type_edit.setReadOnly(True)
        form.addRow(QLabel('Load Type'), type_edit)

        self.variable = self.variables[load_type]
        self.var_edit = QLineEdit()
        form.addRow(QLabel(self.variable), self.var_edit)
        if self.variable == 'Weight':
            self.variable = 'w_' + load_type
        else:
            self.variable = 'n_' + load_type
        val = str(getattr(parent.parent.aircraft.useful_load, self.variable))
        self.var_edit.setText(val)


        self.cg_var = 'cg_' + load_type
        cg_variable = getattr(parent.parent.aircraft.useful_load, self.cg_var)
        self.x_edit = QLineEdit()
        self.x_edit.setText(str(cg_variable[0]))
        x_label = QLabel('X cg')
        x_label.setToolTip('X coordinate at the center of gravity (Ft)')
        form.addRow(x_label, self.x_edit)

        self.y_edit = QLineEdit()
        self.y_edit.setText(str(cg_variable[1]))
        y_label = QLabel('Y cg')
        y_label.setToolTip('Y coordinate at the center of gravity (Ft)')
        form.addRow(y_label, self.y_edit)

        self.z_edit = QLineEdit()
        self.z_edit.setText(str(cg_variable[2]))
        z_label = QLabel('Z cg')
        z_label.setToolTip('Z coordinate at the center of gravity (Ft)')
        form.addRow(z_label, self.z_edit)

        # Buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        form.addWidget(buttons)

    def accept(self):
        # Validate input
        if not self.validate_input():
            return

        cg = [float(self.x_edit.text()), float(self.y_edit.text()), float(self.z_edit.text())]
        var = float(self.var_edit.text())
        setattr(self.parent.parent.aircraft.useful_load, self.cg_var, cg)
        setattr(self.parent.parent.aircraft.useful_load, self.variable, var)

        if hasattr(self.parent.parent.aircraft.mission, self.variable):
            setattr(self.parent.parent.aircraft.mission, self.variable, var)

        self.parent.parent.aircraft.set_weight()
        super().accept()

    def validate_input(self):
        inputs = [self.var_edit, self.x_edit, self.y_edit, self.z_edit]
        valid = True
        for item in inputs:
            try:
                float(item.text())
                item.setStyleSheet("")
            except ValueError:
                item.setStyleSheet("border: 1px solid red;")
                valid = False

        return valid


from PySide6.QtWidgets import QLineEdit, QDialog, QFormLayout, QLabel, QCheckBox, QComboBox, QDialogButtonBox
from PySide6.QtCore import Signal
from ...mission_segments import *

#TODO: make sure these types actually line up
mission_profile_types = {'Takeoff': 'takeoff',
                      'Climb': 'climb',
                      'Cruise': 'cruise',
                      'Descent': 'descent',
                        'Loiter': 'loiter',
                         'Weight Drop': "weight_drop",
                     'Landing': 'landing'}

mission_profile_classes = {'Takeoff': takeoff,
                           'Climb': climb,
                           'Cruise': cruise,
                           'Descent': descent,
                           'Loiter': loiter,
                           'Weight Drop': weight_drop,
                           'Landing': landing
                           }

class mission_profile_edit(QDialog):
    """
    Handles mission profile info
    """
    title_changed = Signal(str)

    variables = {
            'takeoff': {'thrust_setting': 'Thrust Setting', 'time': 'time (s)'},
            'climb': {'start_velocity': 'Start Velocity (ft/s)', 'end_velocity': 'End Velocity (ft/s)',
                      'start_altitude': 'Start Altitude (ft)', 'end_altitude': 'End Altitude (ft)'},
            'cruise': {'mach': 'Mach Number', 'altitude': 'Altitude (ft)', 'set_range': 'Set Range'},
            'descent': {'weight_fraction': 'Weight Fraction'},
            'loiter': {'altitude': 'Altitude', 'time': 'Time', 'mach': 'Mach'},
            'landing': {'weight_fraction': 'Weight Fraction', 'reserve_fuel_fraction': 'Reserve Fuel Fraction'}
        }
    create_segment = False
    index = 0

    def __init__(self, parent, phase):
        self.parent = parent
        self.aircraft = parent.aircraft
        super().__init__(parent)
        self.setWindowTitle(f'Mission Profile Edit - {phase}')

        form = QFormLayout()
        self.setLayout(form)

        phase_edit = QLineEdit()
        phase_edit.setText(phase)
        form.addRow(QLabel('Segment Title'), phase_edit)

        # Find the correct segment object from mission_profile
        segment_obj = None
        i = 0
        for segment in self.aircraft.mission.mission_profile:
            if segment.title == phase:
                self.index = i
                segment_obj = segment
                self.segment = segment
                seg_type = segment.segment_type
                break
            i += 1

        if not segment_obj:
            self.create_segment = True
            seg_type = mission_profile_types[phase]

        # Add a line for segment type
        line_edit = QLineEdit(seg_type.capitalize())
        line_edit.setReadOnly(True)
        form.addRow(QLabel('Segment Type'), line_edit)

        # index_edit
        self.index_edit = QLineEdit(str(self.index))
        form.addRow(QLabel('Index'), self.index_edit)

        self.phase = phase
        self.fields = {}
        self.fields['title'] = phase_edit

        # Dynamically create input fields based on segment type
        for var, label in self.variables[seg_type].items():
            if var == 'set_range':  # Dropdown menu
                self.fields[var] = QCheckBox()
                if not self.create_segment:
                    self.fields[var].setChecked(segment_obj.find_range)
            else:  # Standard numerical input
                self.fields[var] = QLineEdit()
                if not self.create_segment:
                    val = str(getattr(segment_obj, var, ""))
                else:
                    val = None
                self.fields[var].setText(val)

            form.addRow(QLabel(label), self.fields[var])

        # Buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        form.addWidget(buttons)


    def accept(self):
        if not self.validate_input():
            return

        # Save values to mission profile
        input_params = {}
        for var in self.fields:
            if isinstance(self.fields[var], QCheckBox):
                value = self.fields[var].isChecked()
            elif isinstance(self.fields[var], QComboBox):
                value = self.fields[var].currentText()
            else:
                if var == 'title':
                    value = self.fields[var].text()
                else:
                    value = float(self.fields[var].text())
            if not self.create_segment:
                setattr(self.segment, var, value)
            else:
                input_params[var] = value

        if self.create_segment:
            seg_class = mission_profile_classes[self.phase]
            seg = seg_class(**input_params)
            self.aircraft.mission.mission_profile.insert(self.index, seg)
        else:
            if float(self.index_edit.text()) != self.index:
                item = self.aircraft.mission.mission_profile.pop(self.index)
                self.aircraft.mission.mission_profile.insert(int(self.index_edit.text()), item)

        super().accept()  # Close the dialog (if applicable)

    def validate_input(self):
        valid = True
        for var, field in self.fields.items():
            if isinstance(field, QLineEdit):  # Validate numerical inputs
                if var == 'title':
                    try:
                        str(field.text())
                        field.setText(field.text())
                    except ValueError:
                        field.setStyleSheet("border: 1px solid red;")
                        valid = False
                else:
                    try:
                        float(field.text())
                        field.setStyleSheet("")
                    except ValueError:
                        field.setStyleSheet("border: 1px solid red;")
                        valid = False

        try:
            int(self.index_edit.text())
        except ValueError:
            self.index_edit.setStyleSheet("border: 1px solid red;")
            valid = False

        return valid


from .wing import Wing
import numpy as np

class Wing_advanced(Wing):
    input_params = {}
    definition_type = "Sections"
    avl_sections = []
    twist = []
    airfoil = []
    component_type = 'Wing_Advanced'
    sections = []

    def __init__(self, params):
        self.input_params = params
        self.component_type = self.__class__.__name__
        for variable_name, variable_value in params.items():
            if hasattr(self, variable_name.lower()):
                setattr(self, variable_name.lower(), variable_value)
        self.create_sections()

    def create_sections(self):
        # Create Sections
        self.sections = []
        first_section = True

        for section_name, section_params in self.input_params['sections'].items():
            if first_section:
                section_params["xle"] = self.xle
                section_params["yle"] = self.yle
                section_params["zle"] = self.zle
                first_section = False
            else:
                section_params["xle"] = self.sections[-1].xle_tip
                section_params["yle"] = self.sections[-1].yle_tip
                section_params["zle"] = self.sections[-1].zle_tip
                section_params["cr"] = self.sections[-1].ct

            section = wing_section(section_params)
            section.title = section_name
            self.sections.append(section)

        # set variables
        self.span = 0
        self.area = 0
        self.cref = 0
        for sec in self.sections:
            self.span += sec.span
            self.area += sec.area
            self.cref += sec.cref * sec.area
        self.cref *= 1 / self.area
        self.taper = self.sections[-1].ct / self.sections[0].cr
        self.aspect_ratio = self.span**2 / self.area
        self.cr = self.sections[0].cr
        self.ct = self.sections[-1].ct
        self.xle = self.sections[0].xle
        self.yle = self.sections[0].yle
        self.zle = self.sections[0].zle
        self.dihedral = np.arctan((self.sections[-1].zle_tip - self.zle) / (self.span*.5))
        self.sweep_location = 0

        self.sweep = np.arctan((self.sections[-1].xle_tip - self.sections[0].xle)/ (self.span/2))
        self.sweep = np.rad2deg(self.sweep)
        self.set_sweep_angle(self.sweep, self.sweep_location)


        if 'airfoils' in self.input_params:
            self.airfoil = []
            self.twist = []
            for af in self.input_params['airfoils'].values():
                self.twist.append(af['twist'])
                self.airfoil.append(af['airfoil'])

        # Define avl sections
        if not self.twist:
            self.twist = [0] * (len(self.sections) + 1)
        if not self.airfoil:
            self.airfoil = [None] * (len(self.sections) + 1)

        while len(self.airfoil) < len(self.sections):
            self.airfoil.append(self.airfoil[-1])
        while len(self.twist) < len(self.sections):
            self.twist.append(self.twist[-1])

        self._avl_sections = [[self.xle, self.yle, self.zle, self.cr, self.twist[0]]]
        i = 1
        for sec in self.sections[1:]:
            self._avl_sections.append([sec.xle, sec.yle, sec.zle, sec.cr, self.twist[i]])
            i += 1
        self._avl_sections.append([self.sections[-1].xle_tip, self.sections[-1].yle_tip, self.sections[-1].zle_tip, self.ct, self.twist[-1]])


    def set_variables(self):
        # Set variables
        for variable_name, variable_value in self.input_params.items():
            if hasattr(self, variable_name.lower()) and variable_name.lower() != 'sections':
                setattr(self, variable_name.lower(), variable_value)

class wing_section(Wing):
    input_params = {}

    cr = 0
    ct = 0
    span = 0
    sweep = 0
    sweep_le = 0
    sweep_te = 0
    sweep_location = .25
    dihedral = 0

    xle = 0
    yle = 0
    zle = 0
    xle_tip = 0
    yle_tip = 0
    zle_tip = 0

    area = 0
    taper = 0

    def __init__(self, params):
        self.input_params = params
        for param, value in params.items():
            if hasattr(self, param):
                setattr(self, param.lower(), value)
        self.define_geometry()

    def define_geometry(self):
        params_1 = ['sweep_le', 'sweep_te', 'cr', 'span']   # Defined by leading edge and trailing edge sweep angles
        params_2 = ['cr', 'ct', 'span', 'sweep']            # Defined by root and tip chord

        if all (key in self.input_params for key in params_1):
            # section is defined by leading edge and trailing edge sweep angles
            dy = self.span
            dx_le = self.span * np.sin(np.radians(self.sweep_le))
            dx_te = self.span * np.sin(np.radians(self.sweep_te))
            self.ct = self.cr - dx_le + dx_te
            self.taper = self.ct / self.cr
            self.xle_tip = self.xle + dx_le
            self.area = .5 * self.span * (self.ct + self.cr)
        else:
            if 'length' in self.input_params:
                self.input_params['span'] = self.input_params['length'] * 2
                self.length = self.input_params['length']
            else:
                self.length = self.input_params['span'] / 2
            super().__init__(self.input_params)

        self.xle_tip = self.xle + self.length * np.tan(self.sweep_le)
        self.yle_tip = self.yle + self.length
        self.zle_tip = self.zle + self.length * np.tan(self.dihedral)


import os
import subprocess

from .flight_conditions import FlightConditions
from .avl_run import run_AVL, AVL_input, import_coefficients
import numpy as np
import sys

class MissionSegment:
    """
    Base class for all mission_rae2822 segments.
    """
    segment_type = ''
    range = 0
    altitude = 0
    mach = 0
    thrust = 0
    sfc = 0
    velocity = 0
    cl = 0
    cd = 0
    cd0 = 0
    cdw = 0

    fuel_burnt = 0
    weight_fraction = 0
    time = 0
    lift_to_drag = 0
    find_range = False
    run_sim = False
    flight_conditions = []

    wi = 0
    wn = 0

    power_required = 0
    power_required_kw = 0
    input_params = []


    def __init__(self):
        pass

    def breguet_range(self, aircraft, wi):
        return 0, 0, 0, 0, 0


class takeoff(MissionSegment):
    thrust_setting = 100
    input_params = ['thrust_setting', 'time', 'title']

    def __init__(self, thrust_setting=100, time=0, title='takeoff', **kwargs):
        self.title = title
        super().__init__()
        self.time = time * 60
        self.thrust_setting = thrust_setting
        self.segment_type = 'takeoff'

    def breguet_range(self, aircraft, wi=None, wn=None):
        # SFC = fuelflow/thrust
        thrust = self.thrust_setting * aircraft.propulsion.max_thrust
        self.sfc, self.max_thrust = aircraft.propulsion.analyze_performance(
                self.altitude, self.mach, self.thrust
            )
        sfc = self.sfc
        fuel_flow = sfc * thrust  # Lbs / hr
        self.fuel_burnt = fuel_flow * self.time/3600
        self.weight_fraction = (wi - self.fuel_burnt) / wi
        self.thrust = thrust
        self.sfc = sfc
        self.fuel_flow = fuel_flow
        self.wi = wi
        self.wn = wi * self.weight_fraction

class climb(MissionSegment):
    start_velocity = 0
    end_velocity = 0
    start_altitude = 0
    end_altitude = 0
    divisions = 1
    best_climb = False
    power_available = 0
    rate_of_climb = 0
    max_thrust = 0
    K = 0
    input_params = ['start_velocity', 'end_velocity', 'start_altitude', 'end_altitude', 'title']

    def __init__(self, title='climb', start_velocity=0, end_velocity=0, start_altitude=0, end_altitude=0, **kwargs):
        super().__init__()
        self.title = title
        self.__dict__.update(kwargs)
        self.start_velocity = start_velocity
        self.end_velocity = end_velocity
        self.start_altitude = start_altitude
        self.end_altitude = end_altitude
        self.divisions = 1
        self.segment_type='climb'

        self.velocity = .293 * start_velocity + .707 * end_velocity
        self.altitude = .5*(start_altitude + end_altitude)
        fc = FlightConditions(self.altitude, 0)

        self.mach = self.velocity / fc.a
        fc = FlightConditions(self.altitude, self.mach)
        self.flight_conditions = fc

    def breguet_range(self, aircraft, wi=None, wn=None):
        cd0, cdw = aircraft.get_cd0(self.altitude, self.mach)
        self.cd0 = cd0

        if wi:
            weight = wi
        else:
            weight = wn

        if self.run_sim:
            AVL_input(aircraft, aircraft.weight_takeoff, mach=self.mach)
            run_AVL(self.flight_conditions, aircraft, cd0=cd0, cdw=cdw)

            self.cl, self.cd = import_coefficients(aircraft, self)
        else:
            self.cl = aircraft.weight_takeoff / (self.flight_conditions.q * aircraft.sref)* 1.3
            a = aircraft.aero_components['Main Wing'].aspect_ratio
            sweep = aircraft.aero_components['Main Wing'].sweep_le
            if (sweep * 180 / np.pi) > 30:
                e = 4.61 * (1 - .045 * a ** .68) * np.cos(sweep) ** .15 - 3.1
            else:
                e = 1.78 * (1 - .045 * a ** .68) - .64
            cdi = self.cl**2 / (np.pi * a * e)
            self.cd = cd0 + cdi

        self.lift_to_drag = self.cl / self.cd

        K = (self.cd - cd0) / self.cl ** 2
        self.K = K

        g = 32.1
        q = self.flight_conditions.q
        sref = aircraft.sref
        D = self.cd * q * sref
        self.thrust = D
        delta_he = (self.end_altitude + 1 / (2 * g) * self.end_velocity ** 2) - (1 / (2 * g) * self.start_velocity ** 2)

        self.sfc, max_thrust = aircraft.propulsion.analyze_performance(self.altitude, self.flight_conditions.mach, self.thrust)

        self.max_thrust = max_thrust
        self.weight_fraction = np.exp(
        -(self.sfc / 3600) * delta_he / (self.velocity * (1 - D / max_thrust)))
        climb_angle = np.arcsin(max_thrust / weight - D / weight)
        rate_of_climb = self.velocity * np.sin(climb_angle)
        self.rate_of_climb = rate_of_climb
        self.time = (self.end_altitude - self.start_altitude) / rate_of_climb
        self.range = self.velocity * self.time / 6076.12
        self.fuel_burnt = self.sfc * D * self.time
        self.fuel_burnt = self.max_thrust * self.sfc * self.time

        if wn:
            wi = wn / self.weight_fraction
        else:
            wn = wi * self.weight_fraction

        self.wi = wi
        self.wn = wn
        self.fuel_burnt = self.wi - self.wn


class cruise(MissionSegment):
    wn = 0
    range = None
    input_params = ['mach', 'altitude', 'title', 'find_range', 'range']

    def __init__(self, mach=0, altitude=0, title='cruise', find_range=True, range=None, **kwargs):
        self.title = title
        super().__init__()
        self.mach = mach
        self.altitude = altitude
        self.find_range = find_range
        if find_range:
            self.range = range
            if hasattr(self.input_params, 'range'):
                self.input_params.remove('range')
        else:
            self.range = range
        self.run_sim = True
        fc = FlightConditions(self.altitude, self.mach)
        self.flight_conditions = fc
        self.velocity = fc.velocity
        self.segment_type = 'cruise'

    def breguet_range(self, aircraft, wn=None, wi=None):
        # determines the fuel burnt during a set range cruise segment

        self.flight_conditions = aircraft.cruise_conditions

        self.cd0, self.cdw = aircraft.get_cd0(self.altitude, self.mach)
        if self.range is None:
            sys.exit('Please input a desired range or set find_range to true')
            logger.error('Please input a desired range or set find_range to true')

        if wi:
            self.wi = wi
            weight = wi
        elif wn:
            self.wn = wn
            weight = wn

        AVL_input(aircraft, weight)
        run_AVL(self.flight_conditions, aircraft)
        self.cl, self.cd = import_coefficients(aircraft, self)
        self.lift_to_drag = self.cl / self.cd
        self.sfc, max_thrust = aircraft.propulsion.analyze_performance(self.flight_conditions.altitude,
                                                                      self.flight_conditions.mach,
                                                                      self.thrust)

        range_feet = self.range * 6076.12
        self.weight_fraction = np.exp(-range_feet * self.sfc / 3600 / (self.flight_conditions.velocity * self.lift_to_drag))
        if wi:
            self.wn = wi * self.weight_fraction
        elif wn:
            self.wi = wn / self.weight_fraction
        self.fuel_burnt = self.wi - self.wn

        self.time = range_feet / self.velocity


    def set_range(self, aircraft, wi, wn):
        self.wi = wi
        self.wn = wn
        # self.wi = (wi + wn) / 2

        AVL_input(aircraft, self.wi)
        run_AVL(self.flight_conditions, aircraft)
        self.cl, self.cd = import_coefficients(aircraft, self)
        self.lift_to_drag = self.cl / self.cd

        self.weight_fraction = wn / wi
        k = (self.cd - aircraft.cd0) / self.cl ** 2

        g = 32.1
        q = self.flight_conditions.q
        sref = aircraft.sref
        D = self.cd * q * sref
        self.thrust = D
        self.sfc, max_thrust = aircraft.propulsion.analyze_performance(self.flight_conditions.altitude, self.flight_conditions.mach,
                                                      self.thrust)


        sfc = self.sfc / 3600

        self.range = np.log(self.weight_fraction) * self.flight_conditions.velocity * self.lift_to_drag / (-sfc) / 6076.12
        self.fuel_burnt = wi-wn
        range_feet = self.range * 6076.12
        self.time = range_feet / self.flight_conditions.velocity
        self.max_thrust = max_thrust


class descent(MissionSegment):
    input_params = ['title', 'weight_fraction']
    def __init__(self, title='descent', weight_fraction=1, **kwargs):
        self.title = title
        super().__init__()
        self.weight_fraction = weight_fraction
        self.segment_type = 'descent'

    def breguet_range(self, aircraft, wi=None, wn=None):

        if wi:
            self.wi = wi
            self.wn = wi * self.weight_fraction
        else:
            self.wn = wn
            self.wi = wn / self.weight_fraction

        self.power_required = 0
        self.power_required_kw = 0
        self.altitude = .5 * aircraft.cruise_conditions.altitude
        self.velocity = .5 * aircraft.cruise_conditions.velocity
        self.fuel_burnt = self.wi - self.wn
        rate_of_descent = 5000 * 60     # ft/hr
        self.time = self.altitude * 2 / rate_of_descent # TODO make this better
        self.range = self.velocity * .592484 * self.time


class loiter(MissionSegment):
    input_params = ['title', 'altitude', 'time', 'mach']
    def __init__(self, title='loiter', altitude=0, time=0, mach=None, **kwargs):
        """
        Initiate a loiter segment
        :param altitude: Loiter altitude (ft)
        :param time: Loiter time (hours)
        """
        self.title = title
        super().__init__()
        self.altitude = altitude
        self.time = time
        self.segment_type = 'loiter'
        if not mach:
            self.mach = .25

    def breguet_range(self, aircraft, wi=None, wn=None):
        K = 0
        for seg in aircraft.mission.mission_profile:
            if hasattr(seg, 'K'):
                K = seg.K
                self.mach = seg.mach
                break

        if wi:
            weight = wi
        else:
            weight = wn

        cd0, _ = aircraft.get_cd0(self.altitude, self.mach)
        self.cd0 = cd0
        self.flight_conditions = FlightConditions(self.altitude, self.mach)
        #TODO fix this, add an actual K calculator
        self.velocity = np.sqrt(2 * wn / (self.flight_conditions.rho * aircraft.sref) * np.sqrt(K / (3 * cd0)))
        self.mach = self.velocity / self.flight_conditions.a
        self.flight_conditions = FlightConditions(self.altitude, self.mach)
        E = self.time * 3600

        self.cl = weight / (self.flight_conditions.q * aircraft.sref)

        self.cd = cd0 + K * self.cl**2

        self.thrust = self.cd * self.flight_conditions.q * aircraft.sref
        self.sfc, max_thrust = aircraft.propulsion.analyze_performance(self.flight_conditions.altitude,
                                                                      self.flight_conditions.mach,
                                                                      self.thrust)

        self.lift_to_drag = self.cl / self.cd
        self.weight_fraction = np.exp(-E * self.sfc/3600 / self.lift_to_drag)

        if wi:
            wn = wi * self.weight_fraction
        else:
            wi = wn / self.weight_fraction

        self.wn = wn
        self.wi = wi
        self.fuel_burnt = self.wi - self.wn
        self.range = self.time * self.flight_conditions.velocity


class landing(MissionSegment):
    reserve_fuel = 0
    wf_reserve = 0
    w_landing = 0
    input_params = ['title', 'weight_fraction', 'reserve_fuel']

    def __init__(self, title='landing', weight_fraction=1, reserve_fuel=0, **kwargs):
        self.title = title
        super().__init__()
        self.weight_fraction = weight_fraction
        self.wf_reserve = reserve_fuel
        self.segment_type = 'landing'

    def breguet_range(self, aircraft, wi=None, wn=None):
        self.reserve_fuel = aircraft.w_fuel * self.wf_reserve
        w_landing = aircraft.weight_takeoff - aircraft.w_fuel + self.reserve_fuel

        if wi:
            weight = wi
        else:
            weight = wn

        self.wn = w_landing
        self.wi = w_landing / self.weight_fraction
        self.power_required = 0
        self.power_required_kw = 0
        self.fuel_burnt = self.wi - self.wn


class weight_drop(MissionSegment):
    # Instantaneous weight drop
    weight_dropped = 0
    input_params = ['title', 'weight_dropped']

    def __init__(self, title='weight_drop', weight_dropped=0, **kwargs):
        self.title = title
        super().__init__()
        self.weight_dropped = weight_dropped
        self.segment_type = 'weight_drop'

    def breguet_range(self, aircraft, wi=None, wn=None):

        if wi:
            wn = wi - self.weight_dropped
        else:
            wi = wn + self.weight_dropped

        self.wn = wn
        self.wi = wi
        self.fuel_burnt = 0
        self.weight_fraction = 1

#TODO test out wieght drop



from PySide6.QtWidgets import QLineEdit, QLabel, QDialogButtonBox, QWidget, QDialog, QVBoxLayout, QComboBox, \
    QFormLayout, QHBoxLayout
from PySide6.QtCore import Signal

class component_edit(QDialog):
    """ Pop up dialog to edit component parameters"""

    component_info = {"wing": ["Span", "Area", "Taper", "Sweep", "Sweep_location", "Dihedral"],
                      "fuselage": ["Length", "Width", "Height"],
                      "horizontal": ["Span", "Area", "Taper", "Sweep", "Dihedral"],
                      "vertical": ["Span", "Area", "Taper", "Sweep", "Dihedral"],
                      "nacelle": ["Length", "Diameter", "w_engine"]}

    tool_tips = {
        "Span": "Tip to tip span (ft)",
        "Area": "Total planform area (ft^2)",
        "Taper": "Taper ratio, tip chord/root chord",
        "Sweep": "Quarter chord sweep angle (deg)",
        "Sweep_location": "Location of the sweep angle as a percentage of the chord",
        "Dihedral": "Dihedral angle (deg)",
        "Length": "Component total length (ft)",
        "Width": "Component width in the y direction at thickest point (Ft)",
        "Height": "Component height in the z direction at thickest point (Ft)",
        "Diameter": "Component diameter at thickest point (Ft)",
        "Xle": "X value at component leading edge (ft)",
        "Yle": "Y value at component leading edge (ft)",
        "Zle": "Z value at component leading edge (ft)",
        "w_engine": "Engine Weight (lbs)"
    }

    common_parameters = ['Xle', 'Yle', 'Zle']
    for comp, param in component_info.items():
        param.extend(common_parameters)
    component_changed = Signal(str)
    title_changed = Signal(str, str)

    def __init__(self, component, parent, new_component=False):
        # Set Variables, itialize
        self.new_component = new_component
        # TODO Make "Maintain Aspect Ratio" feature
        self.component = component
        self.title = component
        self.changed_variables = {}
        super().__init__(parent)
        self.setWindowTitle("Component Edit")
        aircraft = parent.parent.aircraft
        self.aircraft = aircraft
        if new_component:
            component_type = component.lower()
        else:
            component_type = aircraft.aero_components[component].component_type.lower()
        self.component_type = component_type
        self.layout = QVBoxLayout()

        # Component Type Select
        self.layout.addWidget(QLabel('Component Type'))
        component_type_select = QComboBox()
        component_type_select.setEnabled(False)
        component_type_select.addItems(['Fuselage', 'Wing', 'Horizontal', 'Vertical', 'Nacelle', 'Custom'])
        component_type_select.setCurrentText(component_type.capitalize())
        self.layout.addWidget(component_type_select)

        # Title Edit
        self.layout.addWidget(QLabel('Title'))
        title_edit = QLineEdit()
        if not new_component:
            title_edit.setText(str(component.capitalize()))
        title_edit.textChanged.connect(self.handle_title_changed)
        if component.lower() == 'main wing':
            title_edit.setReadOnly(True)
        self.title_edit = title_edit
        self.layout.addWidget(title_edit)

        # Form for all component info
        form = QFormLayout(self)
        form_widget = QWidget()
        form_widget.setLayout(form)

        # Add all component specific information
        self.input_fields = {}
        self.multi_fields = {}
        self.multi_field_values = {}
        for item in self.component_info[component_type]:

            # Populate line edits
            try:
                if new_component:
                    val = ''
                else:
                    val = getattr(aircraft.aero_components[component], item.lower())
                    if item == 'Sweep' or item.startswith('Dihedral'):
                        val *= 180 / 3.14159
                        val = round(val, 2)
            except KeyError:
                val = ''

            if isinstance(val, list):
                container = QWidget()
                layout = QHBoxLayout(container)
                layout.setContentsMargins(0, 0, 0, 0)
                line_edits = []
                for v in val:
                    line_edit = QLineEdit(str(v))
                    layout.addWidget(line_edit)
                    line_edits.append(line_edit)
                self.multi_fields[item] = line_edits
                self.multi_field_values[item] = val

                for le in line_edits:
                    le.textChanged.connect(lambda _, var=item: self.multi_field_changed(item))

                label = QLabel(item)
                label.setToolTip(self.tool_tips[item])
                form.addRow(label, container)
            else:
                line_edit = QLineEdit()
                line_edit.setText(str(val))
                line_edit.textChanged.connect(lambda text, var=item: self.text_changed(var, text))
                label = QLabel(item)
                label.setToolTip(self.tool_tips[item])
                form.addRow(label, line_edit)
                self.input_fields[item] = line_edit

        self.layout.addWidget(form_widget)

        # Ok and Cancel Buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.layout.addWidget(buttons)

        widget = QWidget()
        widget.setLayout(self.layout)

        self.setLayout(self.layout)

    # Stores each changed variable
    def text_changed(self, var, text):
        if hasattr(self.changed_variables, var):
            try:
                self.changed_variables[var] = float(text)
                self.input_fields[var].setStyleSheet("")
            except:
                self.input_fields[var].setStyleSheet("border: 1px solid red;")
        else:
            try:
                self.changed_variables[var] = float(text)
                self.input_fields[var].setStyleSheet("")
            except:
                self.input_fields[var].setStyleSheet("border: 1px solid red;")

    def multi_field_changed(self, item):
        # validate input
        value = []
        for line in self.multi_fields[item]:
            text = line.text().strip()
            try:
                value.append(float(text))
                line.setStyleSheet("")
            except ValueError:
                line.setStyleSheet("border: 1px solid red;")
                validated = False

        # value = [line.text() for line in self.multi_fields[item]]
        self.multi_field_values[item] = value

    def handle_title_changed(self, title):
        self.title = title
        self.changed_variables['title'] = title

    # Closes popup and edits component parameters, signals parent to update graphics
    def accept(self):
        if not self.validate_input():
            return

        for key, value in self.multi_field_values.items():
            self.changed_variables[key] = value
        arg = []
        for key in self.changed_variables.keys():
            if not key == 'title':
                arg.append((self.component, key.lower(), self.changed_variables[key]))

        if arg:
            if not self.new_component:
                self.aircraft.update_component(arg)
                self.component_changed.emit(self.component)
            else:
                params = {}
                for line in arg:
                    params[line[1]] = line[2]
                params['title'] = self.title
                self.aircraft.add_component(self.component_type, params)
                self.component_changed.emit(self.title)
                self.title_changed.emit('', self.title)
                super().accept()
                return
        else:
            self.component_changed.emit(self.component)

        if 'title' in self.changed_variables:
            title = self.title
            self.aircraft.aero_components[title] = self.aircraft.aero_components[self.component]
            self.aircraft.aero_components.pop(self.component)
            self.aircraft.update_component([(title, 'title', title)])
            self.title_changed.emit(self.component, self.title)
        super().accept()

    def validate_input(self):

        validated = True
        for label, line_edit in self.input_fields.items():
            text = line_edit.text().strip()
            if not text:
                line_edit.setStyleSheet("border: 1px solid red;")

            try:
                float(text)
                line_edit.setStyleSheet("")
            except ValueError:
                line_edit.setStyleSheet("border: 1px solid red;")
                validated = False

        if 'title' in self.changed_variables and self.new_component:

            if self.changed_variables['title'] in self.aircraft.aero_components:
                self.title_edit.setStyleSheet("border: 1px solid red;")
                validated = False

        return validated


from ..component import PhysicalComponent
import numpy as np


class Engine(PhysicalComponent):
    """
    Class containing all parameters and methods related to engine and nacelle component.

    Calculates the weight properties for engine, nacelle, and propulsion subcomponents
    Calculates drag contribution from nacelle
    """

    # Default Values
    aero_body = False
    length = 0          # Nacelle Length (ft)
    diameter = 0        # Nacelle Diameter (ft)
    ln = 4              # inlet to compressor face
    p2 = 30             # max static pressure at compressor face (15-50 psi)
    Wec = 3200          # Weight of
    Q = 1.3             # interference factor
    laminar_percent = .05
    pylon_mounted = True
    thrust_reversal = False
    weight_averages = [1, 0, 0]  # [Raymer, Torenbeek, NASA] - weighted averages used for weight estimation

    _n_engines = 2  # Number of engines
    w_engine = 0    # Weight of each engine
    engine_type = 'turbofan'        # 'Turbofan' or 'Propeller'

    weight_nacelle = 0
    weight_controls = 0
    weight_starter = 0
    weight_fuel_system = 0


    def __init__(self, params):
        """
                Initialize nacelle component and set input parameters.

                params: <dict> list of parameters to edit
                """
        super().__init__(params)
        # Find wetted surface area
        self.s_wet = np.pi * self.diameter * self.length

    def parasite_drag(self, flight_conditions, sref):
        """
        Set nacelle parasite drag using Raymer method.

        :param flight_conditions: flight conditions object at cruise.
        :param float sref: reference area for aircraft.
        """
        l_char = self.length
        f = self.length / self.diameter
        form_factor = 1 + .35/f * self.n_engines     # From Raymer

        super().parasite_drag(form_factor, l_char, flight_conditions, sref)

    def set_weight(self, aircraft, wdg):
        """
        Sets weight of nacelle including all subcomponents

        :param object aircraft: <aircraft object> aircraft which the fuselage belongs to
        :param int wdg: takeoff gross weight (lbs)

        :return: weight of the nacelle
        :rtype: int
        """

        self.weight = 0
        # Set nacelle weight
        self.weight += self.raymer_weight(aircraft, wdg)

        # Add additional turbofan components if required
        if self.engine_type == 'turbofan' or aircraft.aircraft_type == 'transport':
            # Set weight of engines
            self.weight += self.n_engines * self.w_engine
            # Set weight of engine controls
            self.weight += self.controls_weight(aircraft)
            # set weight of starters
            self.weight += self.starter_weight(aircraft)
        # Set weight of fuel systems
        self.weight += self.fuel_system_weight(aircraft)
        self.set_cg()
        return self.weight

    def set_cg(self):
        """
        Set the center of gravity of the nacelle.

        """
        if isinstance(self.xle, list):
            xcg = sum(self.xle) / len(self.xle)
            ycg = sum(self.yle) / len(self.yle)
            zcg = sum(self.zle) / len(self.zle)
        else:
            xcg = self.xle
            ycg = self.yle
            zcg = self.zle
        self.cg = [self.length * .45 + xcg, ycg, zcg]
        super().set_cg()

    def raymer_weight(self, aircraft, wdg):
        """
        Calculates nacelle weight using Raymer method

        :param object aircraft: aircraft object which this nacelle belongs to
        :param int wdg: gross takeoff weight (lbs)

        :return: Raymer weight (lbs)
        :rtype: float
        """

        if self.engine_type == 'turbofan':
            if self.pylon_mounted:
                kng = 1.017
            else:
                kng = 1

            if self.engine_type == 'propeller':
                kp = 1.4
            else:
                kp = 1
            if self.thrust_reversal:
                ktr = 1.18
            else:
                ktr = 1

            wec = 2.331 * self.w_engine ** .901 * kp * ktr
            self.weight_raymer = .6724 * kng * self.length**.1 * self.diameter**.294 * aircraft.ultimate_load**.119 * \
                                  wec ** .611 * self.n_engines**.984 * self.s_wet ** .224
            self.weight_nacelle = self.weight_raymer
        elif self.engine_type == 'propeller':
            self.weight_raymer = self.w_engine ** .922 * self.n_engines
        return self.weight_raymer

    def controls_weight(self, aircraft):
        """
        Sets weight of engine controls

        :param object aircraft: aircraft object which this nacelle belongs to

        :return: weight of the controls (lbs)
        :rtype: int
        """
        if isinstance(self.xle, list):
            xle = sum(self.xle) / len(self.xle)
        else:
            xle = self.xle
        # Raymer method
        lec = aircraft.aero_components['Fuselage'].length - (xle + .5 * self.length)
        self.weight_controls = 5 * self.n_engines + .8 * lec
        return self.weight_controls

    def starter_weight(self, aircraft):
        """
        Calculates weight of engine starters.

        :param object aircraft: aircraft object which this nacelle belongs to

        :return: weight of the engine starters (lbs)
        :rtype: float
        """
        # Raymer method
        w_raymer = 49.19 * (self.n_engines * self.w_engine / 1000) ** .541

        # NASA method
        w_nasa = 11 * self.n_engines * aircraft.cruise_conditions.mach ** .32 * self.diameter

        self.weight_starter = .5 * (w_nasa + w_raymer)
        return self.weight_starter

    def fuel_system_weight(self, aircraft):
        """
        Calculates weight of the fuel system

        :param object aircraft: aircraft object which this nacelle belongs to.

        :return: weight of the fuel system (lbs)
        :rtype: float
        """
        vt = aircraft.mission.w_fuel / aircraft.mission.rho_fuel
        n_tank = aircraft.subsystems.parameters['n_tanks']

        if aircraft.aircraft_type == 'transport':
            # Raymer method
            w_raymer = 1.2025 * vt ** .606 * n_tank ** .5

            # Torenbeek method
            w_torenbeek = 80 * (self.n_engines + n_tank - 1) + 15 * n_tank ** .5 + vt ** .33
            # TODO fix all this to make it accessible to multiple nacelle configurations
            # NASA method
            w_nasa = 1.07 * aircraft.mission.w_fuel ** .58 * n_tank ** .43 * aircraft.cruise_conditions.mach ** .34

            self.weight_fuel_system = (w_nasa + w_raymer + w_torenbeek) / 3

        elif aircraft.aircraft_type == 'general_aviation':
            # Raymer method
            vi = vt  # holding vi/vt = 1 for now (this is the integral tank volume / the total fuel volume
            vi_vt = 1
            w_raymer = 2.49 * vt ** .726 * (1 / (1 + vi_vt)) ** .363 * n_tank ** .242 * self._n_engines ** .157
            # NASA method
            w_nasa = 1.07 * aircraft.mission.w_fuel * .58 * self._n_engines ** .43

            self.weight_fuel_system = (w_nasa + w_raymer)/2

        return self.weight_fuel_system

    @property
    def n_engines(self):
        return (self._n_engines)

    @n_engines.setter
    def n_engines(self, n):
        self._n_engines = n
        n_stations = np.ceil(n / 2).astype(int)

        def le_coords(coord):
            # Make sure coordinate is a list
            if not isinstance(coord, list):
                coord = [coord]
            # If there is only one coordinate, return just that number
            if n_stations == 1:
                return coord[0]
            # Default the recently created engines to 0
            while len(coord) < n_stations:
                coord.append(0)
            return coord

        self.xle = le_coords(self.xle)
        self.yle = le_coords(self.yle)
        if n == 1:
            self.yle = 0
        self.zle = le_coords(self.zle)




class UsefulLoad:
    """
    Contains weight properties all the whole useful load

    crew members, passengers, fuel, and cargo
    """
    weight = 0
    n_pilots = 0
    w_pilots = 0
    cg_pilots = [0, 0, 0]
    n_flight_attendants = 0
    w_flight_attendants = 0
    cg_flight_attendants = [0, 0, 0]
    n_passengers = 0
    w_passengers = 0
    cg_passengers = [0, 0, 0]
    w_fuel = 0
    cg_fuel = [0, 0, 0]
    w_cargo = 0
    cg_cargo = [0, 0, 0]

    cg = [0, 0, 0]
    inertia = [0, 0, 0]


    def __init__(self, params):
        """
        Initialize useful load and set all weights, centers of gravity (cg), and moments of inertia
        :param <dict> params: list of parameters to edit.
        """
        for variable_name, variable_value in params.items():
            if hasattr(self, variable_name.lower()):
                setattr(self, variable_name.lower(), variable_value)

    def set_weight(self, aircraft):
        """
        Intializes and updates the useful load's various weights like cargo and passengers, and uses these to calculate
        the total weight and center of gravity.

        """
        # Set passenger and crew weight
        self.w_passengers = self.n_passengers * 165
        self.w_flight_attendants = self.n_flight_attendants * 165
        self.w_pilots = self.n_pilots * 165
        self.inertia = [0, 0, 0]
        # Update moments of inertia
        if self.cg_passengers:
            self.inertia = [i + self.w_passengers * x for i, x in zip(self.inertia, self.cg_passengers)]

        if self.cg_cargo:
            self.inertia = [i + self.w_cargo * x for i, x in zip(self.inertia, self.cg_cargo)]

        if self.cg_fuel:
            self.inertia = [i + self.w_fuel * x for i, x in zip(self.inertia, self.cg_fuel)]

        if self.cg_pilots:
            self.inertia = [i + self.w_fuel * x for i, x in zip(self.inertia, self.cg_pilots)]

        if self.cg_flight_attendants:
            self.inertia = [i + self.w_fuel * x for i, x in zip(self.inertia, self.cg_flight_attendants)]

        self.w_fuel = aircraft.w_fuel

        self.weight = self.w_pilots + self.w_passengers + self.w_flight_attendants + self.w_fuel + self.w_cargo
        self.cg = [i / self.weight for i in self.inertia]


import logging
import numpy as np

logger = logging.getLogger(__name__)

class Component:
    """
    Super Class for all components

    Contains all relevant component information such as weight, parasite drag, etc.

    All methods may be overriden by specific component classes
    """

    params = {}  # Input parameters, used to edit the component later
    title = ""  # Name of component
    component_type = ""  # Component class type

    weight = 0.0  # Overall weight

    cg = [0, 0, 0]  # Center of gravity (x, y, z)
    inertia = [0, 0, 0]  # Moments of inertia (ix, iy, iz)
    aero_body = False

    def __init__(self, params):
        """
        Initializes component and sets all input parameters

        Parameters:
        :param <dict> params: list of parameters to edit.
        """

        self.params = params
        self.component_type = self.__class__.__name__
        for variable_name, variable_value in params.items():
            if hasattr(self, variable_name.lower()):
                setattr(self, variable_name.lower(), variable_value)

    def update(self, variable, value, maintain_aspect_ratio=True):
        """
        Updates specified parameters in component.
        Note the whole aircraft will need to be re-evaluated to determine weights and drag.

        Parameters:
        :param <str> variable: name of variable to be updated
        :param float value: value to update the variable to

        Returns:
        :return weight: Overall weight.
        :rtype: int
        """
        if hasattr(self, variable):
            setattr(self, variable, value)
        else:
            logging.warning(f'Variable {variable} not found found in list of variables for component {self.__class__}')
        if hasattr(self.params, variable):
            setattr(self.params, variable, value)

    def set_weight(self, aircraft, wdg):
        self.weight = 0
        return self.weight

    def set_cg(self, cg=None):
        """
        Sets center of gravity and moments of inertia for component.

        :param float cg: center of gravity.
        """
        self.cg = cg
        if self.cg:
            self.inertia = [self.weight * x for x in cg]

    def cross_sectional_area(self, aircraft, x):
        pass


class PhysicalComponent(Component):
    """
    Component which effects aerodynamic performance of the aircraft
    """
    # Default Values
    avl_sections = []  # Sections to input into AVL
    attachment = ""  # What component is this attached to, eg. nacelle attached to wing
    cd0 = 0.0  # Parasite drag coefficient
    cdw = 0.0  # Wave drag coefficient
    xle = 0.0  # Leading edge x coordinate
    yle = 0.0  # Leading edge y coordinate
    zle = 0.0  # Leading edge z coordinate

    laminar_percent = 0.1  # Percentage experiencing laminar flow
    Q = 1.0  # Interference factor (for parasite drag)
    s_wet = 0  # Wetted Surface area

    weight_raymer = 0.0  # Weight using Raymer estimation method
    weight_torenbeek = 0.0  # Weight using Torenbeek method
    weight_nasa = 0  # Weight using NASA FLOPS Method
    weight_averages = [1 / 3, 1 / 3, 1 / 3]  # [Raymer, Torenbeek, NASA] - weighted averages used for weight estimation

    def __init__(self, params):
        """
        Initializes the physical_component with the given parameters.

        :param <dict> params: list of parameters to edit
        """
        super().__init__(params)

    def parasite_drag(self, form_factor, l_char, flight_conditions, sref):
        """
        Computes the Parasitic Drag Coefficient using Raymer's method.

        :param float form_factor: Adjusts for component shape effects on drag.
        :param float l_char: Characteristic length of the component.
        :param object flight_conditions: Contains rho (air density), v (velocity), mu (Air dynamic viscosity), and mac (mach number)
        :param float sref: Reference area
        """
        # Parasite drag calculation - From Raymer
        # L is the characteristic length
        rho = flight_conditions.rho
        v = flight_conditions.velocity
        mu = flight_conditions.mu
        mach = flight_conditions.mach
        Re = rho * v * l_char / mu
        cflam = 1.328 / np.sqrt(Re)
        cfturb = .455 / (np.log10(Re) ** 2.58 * (1 + .144 * mach ** 2) ** .65)
        cf = self.laminar_percent * cflam + (1 - self.laminar_percent) * cfturb
        # cf = .0055
        if sref == 0:
            return 0
        else:
            self.cd0 = cf * form_factor * self.Q * self.s_wet / sref

    def set_wave_drag(self, aircraft):
        """
        Sets the wave drag to zero.

        :param object aircraft: Aircraft for which the wave drag is set.

        :return: Zero
        :rtype: int
        """
        self.cdw = 0
        return 0

    def set_weight(self, aircraft, wdg):
        """
        Sets the weight of the component to the weighted averages of the Raymer, Torenbeek and Nasa weight calculations.
        :param object aircraft: Aircraft to which the component belongs to.
        :param int wdg: takeoff Gross Weight.
        :return: Weight of the component.
        :rtype: int
        """
        # exclude torenbeek method for general_aviation aircrafts
        if aircraft.aircraft_type == 'general_aviation':
            self.weight_averages = [.5, 0, .5]

        self.weight = self.raymer_weight(aircraft, wdg) * self.weight_averages[0] + \
                      self.torenbeek_weight(aircraft, wdg) * self.weight_averages[1] + \
                      self.nasa_weight(aircraft, wdg) * self.weight_averages[2]
        self.set_cg()

        return self.weight

    def raymer_weight(self, aircraft, wdg):
        """
        Sets the Raymer weight of the component to zero.

        :param object aircraft: Aircraft to which the component belongs
        :param int wdg: Takeoff gross weight.

        :return: 0
        :rtype: int
        """
        return 0

    def torenbeek_weight(self, aircraft, wdg):
        """"
        Sets the Torenbeek weight of the component to zero.

        :param object aircraft: Aircraft to which the component belongs
        :param int wdg: Takeoff gross weight.

        :return: 0
        :rtype: int
        """
        return 0

    def nasa_weight(self, aircraft, wdg):
        """
        Sets the NASA weight of the component to zero.

        :param object aircraft: Aircraft to which the component belongs
        :param int wdg: Takeoff gross weight.

        :return: 0
        :rtype: int
        """
        return 0

    def set_cg(self):
        """
        Updates the center of gravity of the component.

        """
        super().set_cg(self.cg)


from PySide6.QtWidgets import QWidget, QMainWindow, QHBoxLayout, QFileDialog
from .toolbox.toolbox_main import ToolBox
from ..aircraft import Aircraft
from .graphics import graphics

# git is working


# Subclass QMainWindow to customize your application's main window
class MainWindow(QMainWindow):
    """ Main Window for WUADS """

    def __init__(self, aircraft):
        super().__init__()
        self.setWindowTitle("WUADS")
        self.aircraft = aircraft
        self.initiate_window()

    def initiate_window(self):
        # Initiate widgets and organize layout
        self.resize(1000, 800)

        # Add Toolbox
        layout = QHBoxLayout()
        self.toolbox = ToolBox(self)
        layout.addWidget(self.toolbox, 1)

        # Add graphics window
        self.graphics = graphics(self)
        self.toolbox.component_changed.connect(self.graphics.update_component)
        self.toolbox.component_selected.connect(self.graphics.handleComponentSelected)
        self.toolbox.component_renamed.connect(self.graphics.handleComponentRenamed)
        self.graphics.plot_aircraft(self.aircraft)
        layout.addWidget(self.graphics, 3)

        # Set layout
        container = QWidget()
        container.setLayout(layout)
        self.setCentralWidget(container)

        # Menus
        menu = self.menuBar()
        # File Menu
        file_menu = menu.addMenu('&File')
        load_aircraft = file_menu.addAction('Load')
        load_aircraft.triggered.connect(self.load_config)

        save_aircraft = file_menu.addAction('Save')
        save_aircraft.triggered.connect(self.save_config)

        file_menu.addAction('Close')

    def load_config(self, *args):
        file_dialog = QFileDialog(self)
        file_dialog.setNameFilter("*.yml *.yaml")
        file_dialog.exec()
        config_file = file_dialog.selectedFiles()
        self.aircraft = Aircraft(str(config_file[0]))
        self.initiate_window()

    def save_config(self, *args):
        file_dialog = QFileDialog.getSaveFileName(
            parent=self,
            caption='Save Aircraft Configuration File',
            filter='*.yml *.yaml'
        )
        file_name = file_dialog[0]
        self.aircraft.write_config_file(file_name)



from ..aerobodies.wing import Wing
import numpy as np


class Horizontal(Wing):
    """
    Class containing all methods and parameters related to Horizontal stabilizers.
    Subclass of wing, has all the same variables.
    """
    weight_averages = [.2, .4, .4]  # [Raymer, Torenbeek, NASA] - weighted averages used for weight estimation
    control_surface_ratio = .2

    def __init__(self, params):
        """
        Initializes horizontal stabilizer component

        params: <dict> list of parameters to edit
        """
        super().__init__(params)
        self.Q = 1.1

    def raymer_weight(self, aircraft, wdg):
        """
        Set horizontal stabilizer weight using Raymer method.

        :param object aircraft: <aircraft object> aircraft which the horizontal stabilizer belongs to
        :param int wdg: gross takeoff weight

        :return: Raymer weight (lbs)
        :rtype: float
        """
        if aircraft.aircraft_type == 'transport':
            if self.control_surface_ratio == 1.0:
                kuht = 1.143
            else:
                kuht = 1.0

            lt = (self.xle + .25 * self.cr) - \
                 (aircraft.aero_components['Main Wing'].xle + .25 * aircraft.aero_components['Main Wing'].cr)

            a = .0379 * kuht * (1 + .5 * aircraft.aero_components['Fuselage'].width / self.span) ** -.25
            b = wdg ** .639 * aircraft.ultimate_load ** .1 * self.area ** .75 * lt ** -1
            c = (.3 * lt) ** .704 * np.cos(self.sweep) ** -1 * self.aspect_ratio ** .166
            d = (1 + self.control_surface_ratio) ** .1

            self.weight_raymer = a * b * c * d
            return self.weight_raymer

        elif aircraft.aircraft_type == 'general_aviation':
            a = .016 * (aircraft.ultimate_load * wdg) ** .414
            b = aircraft.cruise_conditions.q ** .168 * self.area ** .896
            c = ((100 * self.tc) / np.cos(self.sweep)) ** -.12
            d = (self.aspect_ratio / (np.cos(self.sweep) ** 2)) ** .043
            e = self.taper ** -.02
            self.weight_raymer = a * b * c * d * e
            return self.weight_raymer



    def torenbeek_weight(self, aircraft, wdg):
        """
        Set horizontal stabilizer weight using Torenbeek Method.

        :param object aircraft: <object aircraft>  aircraft which the horizontal stabilizer belongs to
        :param int wdg: takeoff gross weight (lbs)

        :return: Torenbeek weight (lbs)
        :rtype: float
        """
        if aircraft.aircraft_type == 'transport':
            kh = 1.1

            dive_speed = aircraft.cruise_conditions.velocity * .592484
            a = 1.1 * self.area * kh
            b = 3.81
            c = self.area ** .2 * dive_speed
            d = 1000
            e = np.cos(self.sweep_mid) ** .5
            f = d * e

            self.weight_torenbeek = a * (b * (c / f) - .287)
        # elif aircraft.aircraft_type == 'general_aviation':
        #     a = aircraft.aero_components['Vertical Stabilizer'].area + self.area
        #     self.weight_torenbeek = .04 * (aircraft.ultimate_load * a ** 2) ** .75
        return self.weight_torenbeek

    def nasa_weight(self, aircraft, wdg):
        """
        Calculates horizontal stablilizer weight using NASA FLOPS method.

        :param: object aircraft: <aircraft object> aircraft which the horizontal stabilizer belongs to
        :param: int wdg: takeoff gross weight (lbs)

        :return: NASA FLOPS weight (lbs)
        :rtype: float
        """
        q = aircraft.cruise_conditions.q # 1481.35 * aircraft.mission.mach * aircraft.flight_conditions.delta # cruise dynamic pressure

        if aircraft.aircraft_type == 'transport':
            self.weight_nasa = .53 * self.area * wdg ** .2 * (self.taper + .5)
            return self.weight_nasa
        elif aircraft.aircraft_type == 'general_aviation':
             # cruise dynamic pressure
            self.weight_nasa = .016 * self.area ** .873 * (aircraft.ultimate_load * wdg) ** .414 * q ** .122
            return self.weight_nasa

    def set_wave_drag(self, aircraft):
        """
        Set the wave drag to zero.

        :param object aircraft: <aircraft object> aircraft which the horizontal stabilizer belongs to.
        :return: 0
        :rtype: int
        """
        self.cdw = 0
        return 0


__version__ = "0.1.0"

# Expose core functionality at the package level
from .aircraft import Aircraft
from .mission import Mission
from .propulsion import turbofan, propeller
from .reports import weights_report, mission_profile_report

# Clean up namespace
__all__ = [
    "Aircraft",
    "Mission",
    "turbofan",
    "propeller",
    "weights_report",
    "mission_profile_report",
    "__version__"
]


import pyvista as pv
import numpy as np
from scipy.interpolate import splprep, splev
from PySide6.QtWidgets import QVBoxLayout, QFrame
from pyvistaqt import QtInteractor
from pathlib import Path
from importlib import resources


class graphics(QFrame):
    """
    Main graphics display

    Uses Pyvista for 3D graphics display
    """
    meshes = {}
    selected_component = ''

    # Initiate graphics window widget and plotter
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        # Create frame and graphics plotter
        self.setFrameShape(QFrame.StyledPanel)
        vlayout = QVBoxLayout()
        self.plotter = QtInteractor(self)

        vlayout.addWidget(self.plotter.interactor)
        self.setLayout(vlayout)

    # Plots all components declared in aircraft.aero_components
    def plot_aircraft(self, aircraft):
        # Add meshes for all components
        for comp in aircraft.aero_components.values():
            mesh = plot_component(comp)
            # self.meshes[comp.title] = plot_component(comp)
            try:
                self.meshes[comp.title] = self.plotter.add_mesh(mesh, show_edges=False, color='grey', metallic=.25, pbr=True)
            except:
                print(f'Failed to Mesh {comp.title}')

        # Rotate default camera position
        pos, focal_point, up = self.plotter.camera_position
        new_pos =[-264.1594486764661, -197.83994171675525, 210.9712561359443]
        self.plotter.camera_position = [[-264.1594486764661, -197.83994171675525, 210.9712561359443], focal_point, up]
        self.plotter.reset_camera()
        self.plotter.render()
        # self.plotter.camera.zoom('tight')


    # Updates a given component and replots it
    def update_component(self, component):
        # Remove mesh if the component is deleted
        try:
            comp = self.parent.aircraft.aero_components[component]
        except KeyError:
            if component in self.meshes:
                self.plotter.remove_actor(self.meshes[component])
                self.plotter.update()
            return

        if component in self.meshes:
            self.plotter.remove_actor(self.meshes[component])
        mesh = plot_component(comp)
        self.meshes[component] = self.plotter.add_mesh(mesh, show_edges=False, color='grey', metallic=.25, pbr=True)
        self.plotter.update()


    # Highlights selected component
    def handleComponentSelected(self, component):
        if self.selected_component and self.selected_component in self.meshes:
            self.meshes[self.selected_component].prop.color = 'grey'
        self.selected_component = component
        if component in self.meshes:
            self.meshes[component].prop.color = 'red'
        self.plotter.update()

    def handleComponentRenamed(self, old_title, new_title):
        self.meshes[new_title] = self.meshes.pop(old_title)




def interpolate_curve(x, y, num_points=100):
    # Calculate the spline representation of the curve
    tck, u = splprep([x, y], s=0)

    # Generate a finer parameterization of the curve
    u_interp = np.linspace(0, 1, num_points)

    # Perform spline interpolation to get new x and y values
    x_interp, y_interp = splev(u_interp, tck)

    return x_interp, y_interp

#TODO graphics and gui support for advanced wings

def plot_wing(wing):

    # Load Airfoil Points
    with resources.open_text("WUADS.assets", "naca_0012.dat") as f:
        # Normalize whitespace on each line
        clean_lines = (" ".join(line.strip().split()) for line in f)
        data = np.genfromtxt(clean_lines, delimiter=" ")

    vert = False
    if wing.component_type == 'Vertical':
        vert = True

    # Set profile airfoil
    x_u = data[:len(data) // 2, 0]
    x_l = data[len(data) // 2:, 0]
    y_u = data[:len(data) // 2, 1]
    y_l = data[len(data) // 2:, 1]
    x_profile = np.concatenate([x_u, list(reversed(x_l))])
    z_profile = np.concatenate([y_u, list(reversed(y_l))])

    # Set parameters

    if not vert:
        b = wing.span / 2  # semi span
    else:
        b = wing.span
    taper = wing.taper
    sweep = wing.sweep_le
    if wing.area == 0:
        S = .5 * wing.span * (wing.ct + wing.cr)
    else:
        S = wing.area
    dihedral = wing.dihedral

    x = []
    y = []
    z = []

    n = 50
    if vert:
        yle_ar = np.linspace(wing.zle, wing.zle + b, n)
    else:
        yle_ar = np.linspace(wing.yle, wing.yle + b, n)

    for yle in yle_ar:
        if not vert:
            xle = wing.xle + (yle - wing.yle) * np.tan(sweep)
            zle = wing.zle + (yle - wing.yle) * np.tan(dihedral)
            chord = S / ((1 + taper) * b) * (1 - (1 - taper) / (2 * b) * np.abs(2 * (yle - wing.yle)))
            x.append(np.array(x_profile * chord + xle))
            y.append(np.linspace(yle, yle, len(x_profile)))
            z.append(np.array(z_profile * chord + zle))
        else:
            zle = yle
            xle = wing.xle + (zle - wing.zle) * np.tan(sweep)
            chord = S / ((1 + taper) * b) * (1 - (1 - taper) / (2 * b) * np.abs(2 * (zle - wing.zle))) * 1.3
            x.append(np.array(x_profile * chord + xle))
            y.append(np.array(z_profile*chord))
            z.append(np.linspace(zle, zle, len(x_profile)))

    if vert:
        chord = .001
        x.append(np.array(x_profile * chord + xle))
        y.append(np.array(z_profile * chord))
        z.append(np.linspace(zle, zle, len(x_profile)))
    else:
        choord = .001
        x.append(np.array(x_profile * chord + xle))
        y.append(np.linspace(yle, yle, len(x_profile)))
        z.append(np.array(z_profile * chord + zle))

    return np.array(x), np.array(y), np.array(z)

def plot_fuselage(fuselage):
    if fuselage.diameter == 0 or fuselage.length == 0:
        return None, None, None

    x = [0, 2.84, 6.2, 8.76, 11.67, 17.6, 37.3, 55.5, 81.8, 94.9, 105, 114.7, 125]
    z = [-2.75, -2.17, -1.35, -.62, -.03, 0, 0, 0, 0, .54, 2.13, 3.5, 5.4]
    d = [.1, 5.149, 7.937, 10.3, 12.157, 13.2, 13.2, 13.2, 13.2, 12.120, 8.94, 6.20, 2.4]

    l = fuselage.length
    w = fuselage.diameter

    x_max = x[-1]
    x = [(i / x_max) * l for i in x]
    z = [(i / x_max) * l for i in z]
    d_max = np.max(d)
    d = [i / d_max * w for i in d]
    r = [i * .5 for i in d]

    x_surf = []
    y_surf = []
    z_surf = []

    # Interpolate fuselage curve
    n_length = 50
    n_cross_section = 50
    x_base = x
    x, r = interpolate_curve(x, r, n_length)
    x, z = interpolate_curve(x_base, z, n_length)

    for i in range(len(x)):
        x_fuse = []
        y_fuse = []
        z_fuse = []
        # theta = np.linspace(0, 2 * np.pi, n_cross_section)
        theta = np.linspace(-np.pi / 2, np.pi / 2, n_cross_section)
        for t in theta:
            x_fuse.append(x[i])
            y_fuse.append(r[i] * np.cos(t))
            z_fuse.append(z[i] + r[i] * np.sin(t))
        x_surf.append(np.array(x_fuse))
        y_surf.append(np.array(y_fuse))
        z_surf.append(np.array(z_fuse))

    return np.array(x_surf), np.array(y_surf), np.array(z_surf)

def plot_nacelle(nacelle):
    length = nacelle.length
    diameter = nacelle.diameter
    n=50
    x_arr = np.linspace(nacelle.xle, nacelle.xle+length, n)
    x = []
    y = []
    z = []
    y0 = nacelle.yle
    z0 = nacelle.zle

    for xi in x_arr:
        theta = np.linspace(0, np.pi*2, 30)
        x_cross = []
        y_cross = []
        z_cross = []
        for t in theta:
            x_cross.append(xi)
            y_cross.append(y0 + np.cos(t) * diameter/2)
            z_cross.append(z0 + np.sin(t) * diameter/2)
        x.append(x_cross)
        y.append(y_cross)
        z.append(z_cross)

    return np.array(x), np.array(y), np.array(z)

def plot_component(component):
    if (component.component_type == 'Wing' or
            component.component_type == 'Horizontal' or
            component.component_type == 'Vertical'):
        x, y, z = plot_wing(component)
    elif component.component_type == 'Fuselage':
        x, y, z = plot_fuselage(component)
        if any(v is None for v in (x, y, z)):
            return
    elif component.component_type == 'Engine':

        xle = component.xle
        yle = component.yle
        zle = component.zle
        if not isinstance(xle, list):
            xle = [component.xle]
            yle = [component.yle]
            zle = [component.zle]
        mesh = pv.PolyData()

        stl_path = resources.files("WUADS.assets").joinpath("nacelle.stl")
        for x, y, z in zip(xle, yle, zle):
            # Read STL from package resource path
            temp = pv.read(str(stl_path)).rotate_y(-90, inplace=True)
            bounds = temp.bounds
            xscale = (bounds[1] - bounds[0]) / component.length
            yscale = (bounds[5] - bounds[4]) / component.diameter
            temp = temp.scale([1 / xscale, 1 / yscale, 1 / yscale])
            xyz = [x, y, z]
            temp = temp.translate(xyz)
            mesh = mesh + temp + temp.reflect([0, 1, 0])

        return mesh
    elif component.component_type == 'Wing_Advanced':
        # section 1
        x, y, z = plot_wing(component.sections[0])
        grid = pv.StructuredGrid(x, y, z)
        mesh = grid.extract_surface().triangulate()
        if len(component.sections) > 1:
            for sec in component.sections[1:]:
                x, y, z = plot_wing(sec)
                grid = pv.StructuredGrid(x, y, z)
                mesh1 = grid.extract_surface().triangulate()
                mesh = mesh + mesh1
        mesh = mesh + mesh.reflect((0, 1, 0))
        return mesh
    else:
        return

    grid = pv.StructuredGrid(x, y, z)
    mesh = grid.extract_surface().triangulate()
    mesh = mesh + mesh.reflect((0, 1, 0))

    # plotter.add_mesh(mesh, show_edges=False, color='grey', metallic=.25, pbr=True)
    return mesh



import importlib
import sys
# from ruamel.yaml import YAML
import yaml

from .components.component import Component
from .components.subsystems import Subsystems
from .mission import Mission
from .components.usefulload import UsefulLoad
from .propulsion import turbofan, propeller
from .mission_segments import *
from .flight_conditions import FlightConditions
from .components.aerobodies.wing import Wing
from .components.aerobodies.fuselage import Fuselage
from .components.aerobodies.horizontal import Horizontal
from .components.aerobodies.vertical import Vertical
from .components.aerobodies.engine import Engine
from .components.aerobodies.wing_advanced import Wing_advanced

AEROBODY_CLASSES = {
    "wing": Wing,
    "fuselage": Fuselage,
    "horizontal": Horizontal,
    "vertical": Vertical,
    "engine": Engine,
    "wing_advanced": Wing_advanced,
}

import logging

# Set up basic config: terminal output only
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s'
)
logger = logging.getLogger(__name__)

class Aircraft:
    """
    Class for whole aircraft analysis

    Contains variables and analyses for entire aircraft configuration with input components

    Parameters
    -   Config File: YAML file containing relevant component information for aircraft, see online tutorials for more information
    """

    # Default Values

    title = ''
    aero_components = {}
    cruise_conditions = {}  # Flight conditions at cruise
    sref = 0  # Reference Area (ft^2)
    cd0 = 0  # Parasite Drag coefficient
    cdw = 0  # Wave drag coefficient

    misc_components = {}  # Miscellaneous Components

    cg = [0, 0, 0]
    cg_empty = [0, 0, 0]
    inertia = [0, 0, 0]
    _n_engines = 2

    _lock_component_weights = False  # Locks component weights so editing the aircraft doesn't change them
    _h_cruise = 0  # Cruise Altitude
    _m_cruise = 0  # Cruise Mach number

    weight_takeoff = 0  # Takeoff Gross Weight (lbs)
    weight_empty = 0  # Empty weight, no fuel, no cargo, no crew
    weight_max = 0  # Max Takeoff weight
    weight_reference = 0  # Reference weight used to calculate component weights, typically the same as weight_max

    _w_cargo = 0  # Cargo weight
    _w_fuel = 0  # Fuel Weight
    _n_z = 0  # Ultimate load

    subsystems = []
    useful_load = None
    mission = None
    stability = None
    propulsion = None
    aircraft_type = 'transport'

    _output_dir = None
    _file_prefix = None

    def __init__(self, config_file, wdg_guess=100000):
        """
        Initialize aircraft from config file
        Sets weight and parasite drag

        Parameters:
            config_file: aircraft input file in yaml format, see tutorials for further explanation
        """

        self.input_file = config_file
        self.load_config()
        self.set_weight(wdg_guess=wdg_guess)
        self.set_cd0()
        self.file_prefix = self.title
        if not self._output_dir:
            self.output_dir = os.path.join(os.getcwd(), f'./output/{self.file_prefix}')


    def load_config(self):
        """
        Reads input YAML file and initializes aircraft and declared components
        """

        self.mission = Mission(self)
        with open(self.input_file) as f:
            # yml = YAML(typ='safe', pure=True)
            # config = yml.load(f)
            config = yaml.safe_load(f)

            # Set all aircraft variables
            for variable_name, variable_value in config.get("aircraft", {}).items():
                if hasattr(self, variable_name.lower()):
                    setattr(self, variable_name.lower(), variable_value)
                elif hasattr(self.mission, variable_name.lower()):
                    setattr(self.mission, variable_name.lower(), variable_value)

                if variable_name.lower() == 'mach':
                    self._m_cruise = variable_value
                elif variable_name.lower() == 'altitude':
                    self._h_cruise = variable_value
                elif variable_name.lower() == 'ultimate_load':
                    self._n_z = variable_value

            # Set useful load weights
            self.useful_load = UsefulLoad(config.get("aircraft", {}))

            # Initialize defined aerodynamic components with given parameters
            for component_type, params in config.get("components", {}).items():
                component_class = AEROBODY_CLASSES.get(component_type.lower())
                if 'title' not in params:
                    params['title'] = component_type
                try:
                    self.aero_components[params['title']] = component_class(params)
                except TypeError:
                    logger.warning(f'Component type {component_type} not found, the valid component types are as follows: {AEROBODY_CLASSES.keys()}')

            # Set subsystem parameters for weight estimation
            subsystem_parameters = {}
            for parameter, value in config.get("subsystem_parameters", {}).items():
                subsystem_parameters[parameter] = value
            self.subsystems = Subsystems(subsystem_parameters)

            # general conditions set-up
            self.cruise_conditions = FlightConditions(self.h_cruise, self.mach_cruise)
            self.sref = self.aero_components['Main Wing'].area

            if not 'Main Wing' in self.aero_components:
                raise AttributeError('Main Wing component not declared')

            # Propulsion parameterse

            # try:
            propulsion_parameters = config.get("propulsion", {})

            self.cruise_conditions = FlightConditions(self.h_cruise, self.mach_cruise)
            self.sref = self.aero_components['Main Wing'].area
            self.propulsion = self.generate_propulsion(n_engines=self.n_engines,
                                                       **propulsion_parameters)

            mission_profile_params = config.get("mission_profile", None)
            self.mission.generate_mission_profile(mission_profile_params)

    def generate_propulsion(self, n_engines=None, set_engine=True, **kwargs):
        """
        Sets aircraft propulsion to a turbofan engine with specified values
        :param int n_engines:  Number of engines
        :param float thrust_sea_level:  Maximum available thrust at sea level
        :param float sfc_sea_level:  Specific fuel consumption at sea level
        :param float thrust_cruise:  Maximum available thrust at cruise
        :param sfc_cruise:  Specific fuel consumption at cruise
        """
        if not 'engine_type' in kwargs:
            logger.warn('Engine Type is not declared in propulsion parameters, defaulting to turbofan')
            engine_type = 'turbofan'
        else:
            engine_type = kwargs['engine_type']

        if n_engines:
            self.n_engines = n_engines

        if engine_type.lower() == 'turbofan':
            thrust_sea_level = kwargs.get('thrust_sea_level', None)
            thrust_cruise = kwargs.get('thrust_cruise', None)
            sfc_sea_level = kwargs.get('sfc_sea_level', None)
            sfc_cruise = kwargs.get('sfc_cruise', None)

            engine = turbofan(h_cruise=self.h_cruise,
                              mach_cruise=self.mach_cruise,
                              n_engines=self.n_engines,
                              thrust_sea_level=thrust_sea_level,
                              thrust_cruise=thrust_cruise,
                              sfc_sea_level=sfc_sea_level,
                              sfc_cruise=sfc_cruise)
        elif engine_type.lower() == 'propeller':
            horse_power = kwargs.get('horse_power', None)
            fuel_consumption_rate = kwargs.get('fuel_consumption_rate', None)
            if not horse_power:
                logger.warning('Please enter a base horsepower for propeller engine')
            if not fuel_consumption_rate:
                logger.warning('Please enter a fuel consumption rate for propeller engine')
            engine = propeller(
                n_engines=n_engines,
                horse_power=horse_power,
                fuel_consumption_rate=fuel_consumption_rate
            )


        if set_engine:
            self.propulsion = engine

        return engine

    def set_cd0(self):
        """
        Calculates each components parasite drag coefficient and sets the overall aircraft drag coefficient
        """
        # https://arc.aiaa.org/doi/abs/10.2514/1.47557

        cd0, cdw = self.get_cd0()
        self.cd0 = cd0
        self.cdw = cdw

    def get_cd0(self, height=None, mach=None):
        # https://arc.aiaa.org/doi/abs/10.2514/1.47557

        if height == None:
            height = self.h_cruise
        if mach == None:
            mach = self.mach_cruise

        cd0 = 0
        cdw = 0
        fc = FlightConditions(height, mach)
        for comp in self.aero_components.values():
            comp.parasite_drag(fc, self.sref)
            cd0 += comp.cd0
            cdw += comp.set_wave_drag(self)
        return cd0, cdw

    def set_weight(self, wdg_guess=None, fudge_factor=1.06):
        """
        Uses and iterative loop to set all component weights and overall weight

        Parameters: wdg_guess: Initial guess for gross design weight (lbs)
        fudge_factor: <float> Margin to multiply overall weight by. Good practice to leave at ~1.06 to account
                              for various un modelled weights
        """
        if not wdg_guess:
            wdg_guess = self.weight_takeoff
        self.weight_takeoff = 1
        margin = .00001
        # Start iterative loop
        max_iter = 100

        if self.lock_component_weights:
            wdg_guess = self.weight_max

        for i in range(max_iter):
            # Set structural component weights
            self.weight_takeoff = 0
            for comp in self.aero_components.values():
                self.weight_takeoff += comp.set_weight(self, wdg_guess)
                comp.set_cg()

            for comp in self.misc_components.values():
                self.weight_takeoff += comp.weight

            # Set subsystem Weights
            self.weight_takeoff += self.subsystems.set_subsystem_weights(self, wdg_guess)

            # Add fudge factor
            self.weight_takeoff *= fudge_factor

            # The current weight is the empty weight
            self.weight_empty = self.weight_takeoff
            # Add useful load
            self.useful_load.set_weight(self)
            self.weight_takeoff += self.useful_load.weight
            self.weight_empty += self.useful_load.w_pilots * 1.65 + self.useful_load.w_flight_attendants * 1.65

            # Check if converged
            if np.abs((wdg_guess - self.weight_takeoff) / self.weight_takeoff) < margin or self.lock_component_weights:
                break

            # adjust wdg guess
            wdg_guess = wdg_guess + (self.weight_takeoff - wdg_guess) * .75

        # Set moments of inertia and cg
        self.inertia = [0, 0, 0]
        for comp in self.aero_components.values():
            self.inertia = [i + ix for i, ix in zip(self.inertia, comp.inertia)]
        for comp in self.misc_components.values():
            self.inertia = [i + ix for i, ix in zip(self.inertia, comp.inertia)]

        if not self.lock_component_weights:
            self.weight_max = self.weight_takeoff

        self.inertia = [i + ix for i, ix in zip(self.inertia, self.subsystems.inertia)]
        # Set cg
        self.cg_empty = [i / (self.weight_empty / fudge_factor) for i in self.inertia]

        self.inertia = [i + ix for i, ix in zip(self.inertia, self.useful_load.inertia)]
        self.cg = [i / (self.weight_empty / fudge_factor + self.useful_load.weight) for i in self.inertia]

    def update_component(self, variables, maintain_aspect_ratio=False):
        """
        Updates component and recalculates weight and parasite drag

        Parameters:
            variables: List of variables to update, formatted as tuples with the following format (component, variable, value)
                      example:
                      [('Main Wing', 'Sweep', 25),
                       ('Horizontal', 'Area', 430),
                       ('Fuselage', 'Length', 120)]
            maintain_aspect_ratio: if set to true and the area is altered, the span will be altered to maintain the current
                                   aspect ratio
        """
        # cast variables to list if its just a single tuple
        if not isinstance(variables, list):
            variables = [variables]
        # Set variables
        for var in variables:
            title = var[0]
            variable = var[1]
            value = var[2]
            # Check if title is in aero components
            if title in self.aero_components:
                self.aero_components[title].update(variable, value, maintain_aspect_ratio=maintain_aspect_ratio)

        # Re-initialize the weights and drag
        self.sref = self.aero_components['Main Wing'].area
        self.set_weight()
        self.set_cd0()

    def add_component(self, component_type, params):
        """
        Adds component with a given set of parameters

        component_type: str
        params: dict
        """
        class_name = f"{component_type.capitalize()}"
        module_name = f"WUADS.components.aerobodies.{component_type.lower()}"  # TODO Make this so it can add a non physical component
        module = importlib.import_module(module_name)
        component_class = getattr(module, class_name)
        component = component_class(params)
        self.aero_components[component.title] = component_class(params)
        self.set_weight()
        self.set_cd0()

    def remove_component(self, component):
        """Removes a component and updates parameters"""
        # for key in self.aero_components.keys():
        #     if key.lower() == component.lower():
        #         component = key

        del self.aero_components[component]
        self.set_weight()
        self.set_cd0()

    def write_config_file(self, file_name=None):
        """ Write a .yaml file to save the aircraft's variables """

        if not file_name:
            file_name = os.path.join(self.output_dir, "aircraft.yaml")

        component_list = {}
        for comp in self.aero_components.values():
            params = {}
            for param in comp.params.keys():
                if hasattr(comp, param):
                    val = getattr(comp, param)
                    try:
                        val = float(val)
                    except:
                        pass

                    if param.startswith('sweep') or param.startswith('dihedral'):
                        val *= 180 / np.pi

                    params[param] = val

            component_list[comp.component_type] = params

        subcomp_list = {}
        for comp in self.subsystems.components.values():
            subcomp_list[comp.title] = [float(x) for x in comp.cg]

        subsystem_params = {'subsystems': subcomp_list,
                            'n_nose_wheels': self.subsystems.parameters['n_nose_wheels'],
                            'n_main_wheels': self.subsystems.parameters['n_main_wheels'],
                            'n_tanks': self.subsystems.parameters['n_tanks'],
                            'w_avionics': self.subsystems.parameters['w_avionics']
                            }

        propulsion_params = {'thrust_sea_level': self.propulsion.thrust_sea_level,
                             'thrust_cruise': self.propulsion.thrust_cruise,
                             'sfc_sea_level': self.propulsion.sfc_sea_level,
                             'sfc_cruise': self.propulsion.sfc_cruise}

        mission_profile_params = {}
        for seg in self.mission.mission_profile:
            seg_params = {}
            for item in seg.input_params:
                seg_params[item] = getattr(seg, item)
            mission_profile_params[seg.title] = seg_params

        data = {
            'aircraft': {
                'title': self.title,
                'altitude': self.h_cruise,
                'mach': self.mach_cruise,
                'max_mach': self.mission.max_mach,
                'ultimate_load': self.ultimate_load,
                'w_fuel': self.mission.w_fuel,
                'cg_fuel': self.useful_load.cg_fuel,
                'rho_fuel': self.mission.rho_fuel,
                'n_passengers': self.mission.n_passengers,
                'cg_passengers': self.useful_load.cg_passengers,
                'n_pilots': self.mission.n_pilots,
                'n_flight_attendants': self.mission.n_flight_attendants,
                'cg_crew': self.useful_load.cg_flight_attendants,
                'w_cargo': self.w_cargo,
                'cg_cargo': self.useful_load.cg_cargo,
                'design_range': self.mission.design_range,
                'n_engines': self.n_engines
            },
            'components': component_list,
            'subsystem_parameters': subsystem_params,
            'propulsion': propulsion_params,
            'mission_profile': mission_profile_params

        }

        if not file_name:
            file_name = self.input_file

        with open(file_name, 'w') as file:
            # yaml = YAML(typ='unsafe', pure=True)
            # yaml.default_flow_style = False
            yaml.dump(data, file, default_flow_style=False, sort_keys=False)

    def add_misc_weight(self, title, weight, cg=None):
        """
        Adds a miscellaneous point weight to the aircraft, held in aircraft.misc_components

        :param title: <str> Title of the new component
        :param weight: <float> Weight of the component (lbs)
        :param cg: <[float, float, float]> Component center of gravity, [x, y, z]. Defaults to the cg of the plane if not entered
        """

        if not cg:
            cg = self.cg
        self.misc_components[title] = Component({'title': title, 'weight': weight, 'cg': cg})
        self.set_weight()

    @property
    def w_cargo(self):
        return self._w_cargo

    @w_cargo.setter
    def w_cargo(self, weight):
        """
        Updates the cargo weight of the aircraft to test out performance at different loading conditions
        :param cargo_weight: new cargo weight (lbs)
        """
        self._w_cargo = weight
        if self.weight_max > 0:
            self.useful_load.w_cargo = weight
            self.set_weight()

    @property
    def w_fuel(self):
        return self._w_fuel

    @w_fuel.setter
    def w_fuel(self, weight):
        self._w_fuel = weight
        if self.weight_max > 0:
            self.mission.w_fuel = weight
            self.useful_load.w_fuel = weight
            self.set_weight()

    @property
    def n_engines(self):
        return self._n_engines

    @n_engines.setter
    def n_engines(self, n):
        for comp in self.aero_components.values():
            if comp.component_type == 'Nacelle':
                comp.n_engines = n
        try:
            self.propulsion.n_engines = n
        except AttributeError:
            pass
        self._n_engines = n
        if self.weight_max > 0:
            self.set_weight()

    @property
    def lock_component_weights(self):
        return self._lock_component_weights

    @lock_component_weights.setter
    def lock_component_weights(self, x):
        if x:
            self.weight_max = self.weight_takeoff
        self._lock_component_weights = x

    @property
    def h_cruise(self):
        return self._h_cruise

    @h_cruise.setter
    def h_cruise(self, x):
        self._h_cruise = x
        self.mission.altitude = x
        self.cruise_conditions = FlightConditions(x, self.mach_cruise)
        self.set_cd0()

    @property
    def mach_cruise(self):
        return self._m_cruise

    @mach_cruise.setter
    def mach_cruise(self, x):
        self._m_cruise = x
        self.mission.mach = x
        self.cruise_conditions = FlightConditions(self.h_cruise, x)
        self.set_cd0()

    @property
    def ultimate_load(self):
        return self._n_z

    @ultimate_load.setter
    def ultimate_load(self, nz):
        self._n_z = nz
        if hasattr(self.mission, 'ultimate_load'):
            self.mission.ultimate_load = nz

    @property
    def output_dir(self):
        return self._output_dir

    @output_dir.setter
    def output_dir(self, directory: str):
        # Convert to absolute path (handles both absolute and relative)
        output_dire = os.path.abspath(directory.strip())

        # Create directory if it doesn't exist
        if not os.path.isdir(output_dire):
            try:
                os.makedirs(output_dire)
            except Exception as e:
                raise ValueError(f"Could not create directory: {e}")
        self._output_dir = output_dire

    @property
    def file_prefix(self):
        return self._file_prefix

    @file_prefix.setter
    def file_prefix(self, prefix):
        # check if there's invalid characters in the file prefix
        invalid_chars = r'<>:"/\\|?*'
        if any(char in prefix for char in invalid_chars):
            logger.error('file name contains invalid characters')
            sys.exit(1)
            return False

        # Reserved Windows filenames
        reserved_names = {
            "CON", "PRN", "AUX", "NUL",
            *(f"COM{i}" for i in range(1, 10)),
            *(f"LPT{i}" for i in range(1, 10))
        }
        name_upper = os.path.splitext(prefix)[0].upper()
        if name_upper in reserved_names:
            logger.error('file name is invalid')
            sys.exit(1)
            return False

        # Don't allow path separators
        if os.sep in prefix or (os.altsep and os.altsep in prefix):
            logger.error('file name is invalid, please do not use "/"')
            sys.exit(1)
            return False
        self._file_prefix = prefix



import logging
from ..component import PhysicalComponent
import numpy as np


class Wing(PhysicalComponent):
    """
    Class containing all the variables and method used for a wing component.

    Also serves as the superclass for horizontal and vertical stabilizer.
    """
    # Defualt Values
    sweep = 0           # c/4 sweep (rad)
    sweep_location = .25    # location of the sweep definition (as a percentage of the chord)
    sweep_deg = 0       # c/4 sweep (Deg)
    sweep_le = 0        # Leading edge sweep (rad)
    sweep_mid = 0       # c/2 sweep (rad)
    sweep_te = 0        # Trailing edge sweep (rad)
    sweep_quarter_chord = None
    area = 0            # Planform surface area (ft^2)
    aero_body = True
    span = 0            # Wingspan (ft)
    cref = 0            # Mean Aerodynamic Chord (ft)
    airfoil = ''
    aspect_ratio = 0    # Aspect Ratio
    taper = 0           # Taper Ratio
    cr = 0              # Root Chord
    ct = 0              # Tip Chord
    Q = 1               # interference factor
    laminar_percent = .1    # Percentage of laminar flow
    dihedral = 0        # Dihedral angle (rad)
    dihedral_deg = 0    # Dihedral angle (deg)
    tc = .12            # Airfoil Thickness (At Root)
    hc = .008           # Camber
    weight_averages = [.8, .2, 0]  # [Raymer, Torenbeek, NASA]
    # weight_averages = [.15, .85, 0]
    control_surface_ratio = .1
    cd0 = 0
    laminar_percent = .05
    winglet = {}
    xc = .35    # x value at mean camber line
    avl_sections = []

    def __init__(self, params):
        """
        Initializes the wing with given parameters
        :param dict params: <dict> list of parameters to edit
        """
        super().__init__(params)

        # Check if wing is well-defined
        vars_taper = ['xle', 'yle', 'zle', 'area', 'span', 'sweep', 'dihedral', 'taper']
        vars_chord = ['xle', 'yle', 'zle', 'span', 'cr', 'ct', 'sweep', 'dihedral']
        if all(key in params for key in vars_taper):
            # Wing is defined by taper ratio, define root and tip chord lengths
            cr = 2 * self.area / (self.span * (1 + self.taper))
            self.cr = cr
            ct = cr * self.taper
            self.ct = ct
        elif all(key in params for key in vars_chord):
            # Wing is defined by chord lengths
            ct = self.ct
            cr = self.cr
            area = .5 * (ct + cr) * self.span
            self.area = area
            self.taper = ct / cr
        else:
            logging.warning(" Wing is not well defined")

        # Calculate class variables
        b = self.span / 2
        t = self.taper
        self.cref = cr * (2 / 3) * ((1 + t + t ** 2) / (1 + t))
        self.aspect_ratio = self.span ** 2 / self.area
        xle = self.xle
        yle = self.yle
        zle = self.zle

        self.set_sweep_angle()
        self.dihedral_deg = self.dihedral
        self.dihedral = np.deg2rad(self.dihedral)

        # Create sections used for AVL Analysis
        self.avl_sections = [[xle, yle, zle, self.cr, 0],
                             [(xle + cr / 4) + b * np.tan(self.sweep) - self.ct / 4, b + yle, zle + b * \
                          np.tan(self.dihedral), ct, 0]]

        # Add winglet if exists
        winglet = params.get('winglet', None)
        if winglet is not None:
            section1 = self.avl_sections[1]
            sweep = winglet['sweep']
            height = winglet['height']
            dihedral = winglet['dihedral']
            ct_wl = winglet['ct']
            self.avl_sections.append([section1[0] + np.tan(sweep) * height, section1[1] + height / np.tan(dihedral), section1[2] + height, ct_wl, 0])

        # Set wetted surface area (Torenbeek - advanced aircraft design)
        kq = 0.95
        Qw = kq * self.tc * self.area * np.sqrt(self.area / self.aspect_ratio) / np.sqrt(1 + t)
        self.s_wet = (2 + .5 * self.tc) * (Qw * np.sqrt(self.aspect_ratio * (1 + t)) / (kq * self.tc)) ** (2 / 3)

        # Set airfoils and twist


    def parasite_drag(self, flight_conditions, sref):
        """
        Sets Wing parasite drag coefficient using Raymer Method

        :param flight_conditions: <flight_conditions object> flight conditions at cruise
        :param float sref: reference area for aircraft
        """

        tc = self.tc
        xc = self.xc
        mach = flight_conditions.mach
        sweep = self.sweep
        l_char = self.cref

        # From Raymer
        form_factor = ((1 + tc * .6 / xc + 100 * tc ** 4) * (1.34 * mach ** .18 * (np.cos(sweep)) ** .28))
        # TODO Add shevell method
        super().parasite_drag(form_factor, l_char, flight_conditions, sref)

    def set_wave_drag(self, aircraft):
        """
        Set wave drag for the wing
        Uses methods from Gur and Mason
        https://doi.org/10.2514/1.47557

        :param object aircraft: aircraft object which this wing belongs to.

        :return: Wave drag for the wing
        :rtype: int
        """
        # split wing into strips
        n_strips = 20
        chords = np.linspace(self.cr, self.ct, n_strips+1)
        dy = .5 * self.span / n_strips
        rho = aircraft.cruise_conditions.rho
        v = aircraft.cruise_conditions.velocity
        m = aircraft.cruise_conditions.mach
        if m < .5 or aircraft.sref == 0:
            return 0

        # Eliptical lift distribution equations
        lift = aircraft.weight_takeoff * .99
        gamma_0 = lift / (.25 * np.pi * v * rho * self.span)
        y_dist = np.linspace(self.yle, self.yle + self.span/2, n_strips)

        cdw = 0
        for i in range(n_strips):
            cr_i = chords[i]
            ct_i = chords[i+1]
            c = .5 * (cr_i + ct_i)
            area = .5 * (cr_i + ct_i) * dy

            # Elliptical lift dist
            gamma = gamma_0 * np.sqrt(1 - (2 * y_dist[i] / self.span) ** 2)
            lprime = rho * v * gamma
            cl = lprime / (.5 * rho * v ** 2 * c)
            # Find the drag divergence number
            ka = .95        # TODO make an input to change the airfoil type
            cos = np.cos(self.sweep_mid)
            mdd = (ka - cl / (10 * cos ** 2) - self.tc / cos) / cos
            mcr = mdd - .1077217

            # Set section wave drag coefficient if applicable
            if m > mcr:
                cdw += 20 * (m - mcr) ** 4 * (area / aircraft.sref)

        self.cdw = cdw
        return cdw

    def raymer_weight(self, aircraft, wdg):
        """
        Calculate wing weight using Raymer's method

        :param aircraft: <aircraft object> aircraft which the wing component belongs to
        :param float wdg: design gross weight of aircraft

        :return: Raymer weight
        :rtype: float
        """
        if aircraft.aircraft_type == 'transport':
            scsw = self.control_surface_ratio * self.area
            self.weight_raymer = .0051 * (
                        wdg * aircraft.ultimate_load) ** .557 * self.area ** .649 * self.aspect_ratio ** .5 \
                                 * self.tc ** -.4 * (1 + self.taper) ** .1 * np.cos(self.sweep) ** -1 * scsw ** .1
        elif aircraft.aircraft_type == 'general_aviation':
            W_fw = aircraft.w_fuel # weight of fuel in wing
            a = .036 * self.area ** .758 * W_fw ** .0035
            b = (self.aspect_ratio / (np.cos(self.sweep))**2 ) ** .6
            c = aircraft.cruise_conditions.q ** .006 * self.taper ** .04
            d = ((100 * self.tc) / (np.cos(self.sweep)))**-.3
            e =  (aircraft.ultimate_load * wdg) ** .49

            self.weight_raymer = a * b * c * d * e

        return self.weight_raymer

    def torenbeek_weight(self, aircraft, wdg):
        """
        Calculate wing weight using Torenbeek method

        :param object aircraft: <object aircraft>  aircraft which the wing belongs to
        :param int wdg: takeoff gross weight (lbs)

        :return: Torenbeek weight (lbs)
        :rtype: float
        """
        # only works for transport aircraft
        sweep = self.sweep_mid
        wmzf = wdg - aircraft.mission.w_fuel

        a = .0017 * wmzf * (self.span / np.cos(sweep)) ** .75
        b = 1 + (6.3 * np.cos(sweep) / self.span) ** .5
        c = aircraft.ultimate_load ** .55
        d = (self.span * self.area / (self.cr * self.tc * wmzf * np.cos(sweep))) ** .3
        self.weight_torenbeek = a * b * c * d
        return self.weight_torenbeek

    # NASA estimation method doesn't work for this

    def set_cg(self):
        """
        Sets center of gravity and moments of inertia for wing
        """
        a = self.ct
        b = self.cr
        semi_span = self.span * .5
        c = semi_span * np.tan(self.sweep_le)
        dihedral = self.dihedral

        self.cg = [self.xle + (2*a*c + a**2 + c*b + a*b + b**2) / (3*(a+b)),
                   self.yle + (semi_span * (2*a+b) / (3*(a+b))),
                   self.zle + (semi_span * (2*a+b) / (3*(a+b)))*np.tan(dihedral)]

        super().set_cg()

    def update(self, variable, value, maintain_aspect_ratio=True):
        """
        Updates the wing parameters. Note the whole aircraft will need to be re-evaulated to determine weights and drag

        :param string variable: name of variable to be updated
        :param value: value to update the variable to
        :param boolean maintain_aspect_ratio: if set to true and the area is altered, the span will be altered to maintain the current
                                   aspect ratio
        """

        vars_taper = ['xle', 'yle', 'zle', 'area', 'span', 'sweep', 'dihedral', 'taper']
        vars_chord = ['xle', 'yle', 'zle', 'span', 'cr', 'ct', 'sweep', 'dihedral']
        if (variable == 'area' or variable == 'taper') and all(key in self.params for key in vars_chord):
            self.params.pop('cr')
            self.params.pop('ct')
            self.params['area'] = self.area
            self.params['taper'] = self.ct / self.cr
            return

        # Update span to maintain aspect ratio if applicable
        if maintain_aspect_ratio and variable == 'area':
            self.params['span'] = np.sqrt(self.aspect_ratio * value)
        if maintain_aspect_ratio and variable == 'span':
            self.params['area'] = value ** 2 / self.aspect_ratio

        variable = variable.lower().strip()

        self.params[variable] = value
        self.__init__(self.params)

    def set_sweep_angle(self, sweep=None, sweep_location=None):
        """
        Sets the sweep angle of the wing
        :param sweep: sweep angle in degrees
        :param sweep_location: percentage of the chord at which the sweep is defined (0 for leading edge, .25 for quarter chord, etc.)
        :return:
        """
        if not sweep:
            sweep = self.sweep
        if not sweep_location:
            sweep_location = self.sweep_location

        self.sweep_location = sweep_location
        self.sweep_deg = sweep
        self.sweep = np.deg2rad(self.sweep)

        xle = self.xle
        x_root = xle + sweep_location * self.cr
        b = self.span / 2
        x_tip = x_root + b * np.tan(self.sweep)
        xle_tip = x_tip - sweep_location * self.ct
        self.sweep_le = np.arctan((xle_tip - xle) / b)
        self.sweep_quarter_chord = np.arctan(((xle_tip + .25 * self.ct) - (xle + .25 * self.cr)) / b)
        self.sweep_mid = np.arctan(((xle_tip + .5 * self.ct) - (xle + .5 * self.cr)) / b)
        self.sweep_te = np.arctan(((xle_tip + self.ct) - (xle + self.cr)) / b)




import logging

from .component import Component
from .aerobodies.wing import Wing
import numpy as np
logger = logging.getLogger(__name__)

# Contains component classes for all various subsystems
class Subsystems:
    """
    Contains weight properties for all supported subsystems
    """
    weight = 0
    cg = 0
    inertia = [0, 0, 0]

    parameters = {}
    components = {}

    def __init__(self, params):
        """
        Initializes all declared subsystems and sets parameters

        :param <dict> params: list of parameters to edit
        """
        self.parameters = params
        for name, cg in params['subsystems'].items():
            self.components[name] = Component({'title': name, 'cg': cg})

        # TODO take this out
        if not 'w_avionics' in self.parameters:
            self.parameters['w_avionics'] = 1200

    def set_subsystem_weights(self, aircraft, wdg):
        """
        Calculates weight, cg, and moments of inertia for all subsystems

        :param object aircraft: the aircraft which the weight, cg, and inertia calculations are done.

        :return weight of subsystems
        :rtype float
        """
        self.weight = 0
        self.inertia = [0, 0, 0]
        # Generate all components
        for name, comp in self.components.items():
            name = name.lower()

            if hasattr(self, name) and callable((getattr(self, name))):
                fun = getattr(self, name)
                self.weight += fun(aircraft, wdg, comp)
            else:
                logging.warning(f'Subsystem {name} does not exist')

            # Set inertia value, add to overall inertia
            comp.inertia = [comp.weight * x for x in comp.cg]
            self.inertia = [i1 + i2 for i1, i2 in zip(self.inertia, comp.inertia)]
        self.cg = [i / self.weight for i in self.inertia]
        return self.weight

    def nose_landing_gear(self, aircraft, wdg, comp):
        """
        Calculates weight of nose landing gear

        :param object aircraft: aircraft which the nose landing gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for nose landing gear.

        :return component weight
        :rtype float

        """
        if aircraft.aircraft_type == 'transport':
            params = self.parameters
            # Raymer method (transport)
            if 'knp' in params:
                knp = params['knp']
            else:
                knp = 1
            w_raymer = .032 * knp * wdg ** .646 * aircraft.ultimate_load ** .2 * comp.cg[0] ** .5 * \
                       params['n_nose_wheels']

            # Torenbeek method (transport)
            if aircraft.aero_components['Main Wing'].zle > 0:
                kgr = 1.08
            else:
                kgr = 1.0
            ag = 25
            bg = 0
            cg = .0024
            dg = 0
            w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)

            comp.weight = .5 * (w_raymer + w_torenbeek)

        elif aircraft.aircraft_type == 'general_aviation':
            # Raymer
            w_raymer = .125 * (aircraft.ultimate_load * wdg) ** .566 * (comp.cg[0] / 12) ** .845
            # Torenbeek
            # if aircraft.aero_components['Main Wing'].zle > 0:
            #     kgr = 1.08
            # else:
            #     kgr = 1.0
            # ag = 25
            # bg = 0
            # cg = .0024
            # dg = 0
            # w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
            comp.weight = w_raymer  #* .2 + w_torenbeek * .8




        comp.inertia = [comp.weight * x for x in comp.cg]
        self.components[comp.title] = comp
        return comp.weight

    def main_landing_gear(self, aircraft, wdg, comp):
        """
        Sets main landing gear weight

        :param object aircraft: aircraft which the main landing gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for main landing gear.

        :return component weight
        :rtype float
        """
        params = self.parameters
        if aircraft.aircraft_type == 'transport':
            # Torenbeek method
            if aircraft.aero_components['Main Wing'].zle > 0:
                kgr = 1.08
            else:
                kgr = 1.0
            ag = 20
            bg = .1
            cg = .019
            dg = 0
            w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
            comp.weight = w_torenbeek

        elif aircraft.aircraft_type == 'general_aviation':
            ultimate_load_landing = aircraft.ultimate_load
            w_mlg = wdg  # landing weight
            length_mlg = comp.cg[0] # landing gear length (fix?)
            # raymer method
            w_raymer = .095 * (ultimate_load_landing * w_mlg) ** .768 * (length_mlg / 12) ** .409
            # NASA method
            w_nasa = .0117 * w_mlg ** .95 * length_mlg

           # torenbeek (need to make sure constants are correct)
            #if aircraft.aero_components['Main Wing'].zle > 0:
            #     kgr = 1.08
            # else:
            #     kgr = 1.0
            # ag = 20
            # bg = .1
            # cg = .019
            # dg = 0
            # w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
            # print(w_torenbeek)
            #comp.weight = (w_raymer* .2 + w_nasa * .2 + w_torenbeek * .6)
            #comp.weight =( w_raymer + w_nasa ) / 2
            comp.weight = w_raymer

        return comp.weight

    def air_conditioning(self, aircraft, wdg, comp):
        """
        Calculates weight of air conditioning
        """
        # Uses nasa flops method (same for general and transport)
        l = aircraft.aero_components['Fuselage'].length
        w = aircraft.aero_components['Fuselage'].width
        h = aircraft.aero_components['Fuselage'].height
        w_nasa = 3.2 * (l * w * h) ** .6 + (9 * aircraft.mission.n_passengers ** .83) * aircraft.mission.max_mach + \
                 .075 * self.parameters['w_avionics']
        comp.weight = w_nasa
        if aircraft.aircraft_type == 'general_aviation':
             #Raymer method (experimental)
             w_raymer = .7 * (.265 * wdg ** .52 * aircraft.mission.n_passengers ** .68 * self.parameters['w_avionics'] ** .17 * aircraft.mission.max_mach ** .08)

             comp.weight = .5 * (w_raymer + w_nasa)
        return comp.weight

    def anti_ice(self, aircraft, wdg, comp):
        """
        Calculate weight of anti ice components

        :param object aircraft: aircraft which the anti-ice gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for anit-ice gear.

        :return component weight
        :rtype float
        """
        # Raymer method
        w_raymer = .002 * wdg

        # find nacelle group
        for compon in aircraft.aero_components.values():
            if compon.component_type == 'Nacelle':
                nacelle = compon

        try:
            d = nacelle.diameter * 3 / 5
        except:
            return 0

        # Nasa flops method
        b = aircraft.aero_components['Main Wing'].span
        sweep = aircraft.aero_components['Main Wing'].sweep
        l = aircraft.aero_components['Fuselage'].length
        w_nasa = b / np.cos(sweep) + (3.8 * d * aircraft.n_engines * 3) + 1.5 * l

        comp.weight = .5 * (w_nasa + w_raymer)
        return comp.weight

    def apu(self, aircraft, wdg, comp):
        """
        Calculate auxiliary power unit weight using torenbeek and nasa methods.

        :param object aircraft: aircraft  which the apu belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the apu.

        :return component weight
        :rtype float
        """
        if aircraft.aircraft_type == 'transport':
            # Torenbeek method
            w_torenbeek = .009 * wdg

            # nasa flops method
            d = aircraft.aero_components['Fuselage'].diameter
            l = aircraft.aero_components['Fuselage'].length
            w_nasa = 54 * (d * l) ** .3 + 5.4 * aircraft.mission.n_passengers ** .9

            comp.weight = .5 * (w_nasa + w_torenbeek)
            return comp.weight
        else:
            comp.weight = 0
            return 0

    def avionics(self, aircraft, wdg, comp):
        """
        Set avionics weight using raymer, torenbeek, and nasa methods

        :param object aircraft: aircraft which the avionics belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the avionics.

        :return component weight
        :rtype float
        """
        if aircraft.aircraft_type == 'transport':
            # Raymer method
            w_raymer = 1.73 * self.parameters['w_avionics'] ** .983

            # Torenbeek method
            w_torenbeek = .575 * (wdg - aircraft.mission.w_fuel) ** .556 * aircraft.mission.design_range ** .25

            # nasa method
            d = aircraft.aero_components['Fuselage'].diameter
            l = aircraft.aero_components['Fuselage'].length
            w_nasa = 15.8 * aircraft.mission.design_range ** .1 * aircraft.mission.n_pilots ** .7 * (l * d) ** .43

            comp.weight = (w_nasa + w_torenbeek + w_raymer) / 3


        elif aircraft.aircraft_type == 'general_aviation':
            # Raymer method
            w_raymer = 2.117 * self.parameters['w_avionics'] ** .933

            # # nasa method (same as transport)
            # d = aircraft.aero_components['Fuselage'].diameter
            # l = aircraft.aero_components['Fuselage'].length
            # w_nasa = 15.8 * aircraft.mission.design_range ** .1 * aircraft.mission.n_pilots ** .7 * (l * d) ** .43
            #comp.weight = (w_raymer + w_nasa) / 2
            comp.weight = w_raymer


        return comp.weight

    def electronics(self, aircraft, wdg, comp):
        """
        Calculates weight of all electronics on aircraft

        :param object aircraft: aircraft which the electronic belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the electronics.

        :return component weight
        :rtype float
        """
        # Only uses nasa method (same for both general aviation and transport aircrafts)
        if aircraft.aircraft_type == 'transport':
            d = aircraft.aero_components['Fuselage'].diameter
            l = aircraft.aero_components['Fuselage'].length
            nflcr = aircraft.mission.n_pilots + aircraft.mission.n_flight_attendants
            npass = aircraft.mission.n_passengers
            w_nasa = 92 * l ** .4 * d ** .14 * aircraft.n_engines ** .69 * (1 + .044 * nflcr + .0015 * npass)
            comp.weight = w_nasa

        elif aircraft.aircraft_type == 'general_aviation':
            # raymer
            w_avionics = self.parameters['w_avionics']
            w_fs = aircraft.aero_components['Nacelle'].weight_fuel_system
            comp.weight = 12.57 * (w_fs + w_avionics)**.51
        return comp.weight

    def flight_controls(self, aircraft, wdg, comp):
        """
        Calculates weight of flight controls.

        :param object aircraft: aircraft which the flight controls belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the flight controls.

        :return component weight
        :rtype float
        """


        scs = 0 # Control surface area
        scs_i = 0   # used to find center of gravity
        for compon in aircraft.aero_components.values():
            if isinstance(compon, Wing):
                scs += compon.area * compon.control_surface_ratio
                scs_i += compon.area * compon.control_surface_ratio * \
                         (compon.xle + .25* compon.span * np.tan(compon.sweep))

        comp.cg = [scs_i / scs, 0, 0]

        # Raymer is innacurate
        # Torenbeek method (only factors in for transport aircrafts)
        w_torenbeek = .64 * wdg ** (2/3)

        # Nasa method

        if aircraft.aircraft_type == 'transport':
            w_nasa = 1.1 * aircraft.cruise_conditions.mach ** .52 * scs ** .6 * wdg ** .32
            comp.weight = .5 * (w_nasa + w_torenbeek)
        elif aircraft.aircraft_type == 'general_aviation':
            t0 = 519  # Rankine constant
            L = 0.003575  # temp lapse rate (R/ft)
            R = 1716.5  # Gas Constant
            t = t0 - L * aircraft.cruise_conditions.h
            theta = t / t0
            delta = theta ** (32.2 / (L * R))
            q_dive = 1481.35 * delta * aircraft.mission.max_mach ** 2
            w_nasa = .404 * aircraft.aero_components['Main Wing'].area ** .317 * (wdg / 1000) ** .602 * aircraft.ultimate_load ** .525 * q_dive ** .345
            comp.weight = w_nasa


        return comp.weight

    def furnishings(self, aircraft, wdg, comp):
        """
        Calculates weight of all miscellaneous furnishings.

        :param object aircraft: aircraft which the furnishings belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the furnishings.

        :return component weight
        :rtype float
        """

        if aircraft.aircraft_type == 'transport':
            # Only using nasa, raymer and torenbeek are both complex and inacurate (nasa is same for transport and gen. av.)
            nflcr = aircraft.mission.n_pilots = aircraft.mission.n_flight_attendants
            npass = aircraft.mission.n_passengers
            w = aircraft.aero_components['Fuselage'].width
            d = aircraft.aero_components['Fuselage'].diameter
            xlp = aircraft.aero_components['Fuselage'].length * 80 / 125
            comp.weight = 127 * nflcr + 44 * npass + 2.6 * xlp * (w + d)
        elif aircraft.aircraft_type == 'general_aviation':
            comp.weight = .0582 * wdg - 65
        return comp.weight

    def hydraulics(self, aircraft, wdg, comp):
        """
        Calculates weight of hydraulics.

        :param object aircraft: aircraft which the hydraulics belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the hydraulics.

        :return component weight
        :rtype float
        """
        # Raymer is bad for this one

        # Torenbeek method
        w_torenbeek = .012 * wdg

        # nasa method
        d = aircraft.aero_components['Fuselage'].diameter
        l = aircraft.aero_components['Fuselage'].length

        b = aircraft.aero_components['Main Wing'].span
        s = aircraft.aero_components['Main Wing'].area

        mach = aircraft.cruise_conditions.mach

        n_fuse_engine = 0
        n_wing_engine = 0
        for compo in aircraft.aero_components.values():
            if compo.component_type == 'Nacelle':
                if compo.attachment == 'Main Wing':
                    n_wing_engine += 2
                elif compo.attachment == 'Fuselage':
                    n_fuse_engine += 2

        w_nasa = .57 * ((d * l) + (.27 * s)) * (1 + .03 * n_wing_engine + .05 * n_fuse_engine) * mach ** .33

        comp.weight = .5 * (w_nasa + w_torenbeek)

        # this is what was here before but im not sure where I got it from
        # elif aircraft.aircraft_type == 'general_aviation':
        #     s_w = aircraft.aero_components['Main Wing'].area
        #     s_vt = aircraft.aero_components['Vertical Stabilizer'].area
        #     s_ht = aircraft.aero_components['Horizontal Stabilizer'].area
        #     h_tc = aircraft.aero_components['Horizontal Stabilizer'].tc
        #
        #     w_nasa = .6053 * (s_w + 1.44 * (s_ht/(2 + .387 * h_tc)+ s_vt))
        #
        #     comp.weight = w_nasa

        return comp.weight

    def instruments(self, aircraft, wdg, comp):
        """
        Calculates instruments weights.

        :param object aircraft: aircraft which the instruments belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the instruments.

        :return component weight
        :rtype float
        """
        # Raymer method
        kr = 1.0
        ktp = 1.0
        l = aircraft.aero_components['Fuselage'].length
        d = aircraft.aero_components['Fuselage'].diameter
        b = aircraft.aero_components['Main Wing'].span
        nflcr = aircraft.mission.n_pilots
        nen = aircraft.n_engines
        w_raymer = 4.509 * kr * ktp * nflcr ** .541 * nen * (l + b) ** .5

        # torenbeek method
        w_torenbeek = nflcr * (15 + .032 * (wdg / 1000)) + nen * (5 + .006 * (wdg / 1000))

        # nasa method
        mach = aircraft.cruise_conditions.mach
        w_nasa = .48 * (l * d) ** .57 * mach ** .5 * (10 + 2.5 * nflcr + nen)

        if aircraft.aircraft_type == 'transport':
            comp.weight = (w_nasa + w_torenbeek + w_raymer) / 3
        elif aircraft.aircraft_type == 'general_aviation':
            comp.weight = w_nasa

        return comp.weight